from copy import deepcopy
from typing import Any, Tuple, Union, cast

from astropy.coordinates import Angle, SkyCoord
from astropy.visualization.wcsaxes import WCSAxes
from astropy.wcs import WCS
from matplotlib.patches import Circle

from imephu.annotation import Annotation
from imephu.geometry import pixel_scales, rotate, sky_position_to_pixel, translate


class CircleAnnotation(Annotation):
    """An annotation for adding a circle to a plot.

    The radius of the circle is given as an angular distance on the sky, in the
    direction of the right ascension. A circle is drawn irrespective of the pixel scales
    for right ascension and declination. Hence the circle in general is not a circle on
    the sky but just on the finder chart.

    Parameters
    ----------
    center: `~astropy.coordinates.SkyCoord`
        The right ascension and declination of the circle's center.
    radius: `~astropy.coordinates.Angle`
        The radius of the circle, as an angular distance on the sky in right ascension
        direction.
    wcs: `~astropy.wcs.WCS`
        WCS object.
    edgecolor: color, default: "black"
        The edge color.
    facecolor: color, default: "none"
        The filling color.
    **kwargs: dict, optional
        Additional keyword arguments, which will be passed to Matplotlib's
        `~matplotlib.patches.Circle` patch constructor.
    """

    def __init__(
        self,
        center: SkyCoord,
        radius: Angle,
        wcs: WCS,
        edgecolor: Union[str, Tuple[float, float, float]] = "black",
        facecolor: Union[str, Tuple[float, float, float]] = "none",
        **kwargs: Any,
    ):
        self._center = center
        self._radius = radius
        self._wcs = wcs
        self._kwargs = deepcopy(kwargs)
        self._kwargs["edgecolor"] = edgecolor
        self._kwargs["facecolor"] = facecolor

    def add_to(self, ax: WCSAxes) -> None:
        """Add the circle to a finder chart.

        ax: `~astropy.visualization.wcsaxes.WCSAxes`
            WCS axes object.
        """
        center_px = cast(
            Tuple[float, float], sky_position_to_pixel(self._center, self._wcs)
        )
        ra_pixel_scale, _ = pixel_scales(self._wcs)
        radius_px = float(self._radius / ra_pixel_scale)
        circle_patch = Circle(center_px, radius_px, **self._kwargs)
        ax.add_patch(circle_patch)

    def rotate(self, pivot: SkyCoord, angle: Angle) -> "CircleAnnotation":
        """Rotate this annotation around a pivot and return the result.

        The rotation angle is an angle on the sky, measured from north to east.

        Parameters
        ----------
        pivot: `~astropy.coordinates.SkyCoord`
            Point around which to rotate the annotation.
        angle: `~astropy.coordinates.Angle`
            Angle of rotation, measured from north to east.

        Returns
        -------
        `~imephu.annotation.general.CircleAnnotation`
            The annotation resulting from the rotation.
        """
        rotated_annotation = deepcopy(self)
        rotated_center = rotate(self._center, pivot, angle, self._wcs)
        rotated_annotation._center = rotated_center
        return rotated_annotation

    def translate(self, displacement: Angle) -> "CircleAnnotation":
        """
        Move this annotation along a displacement vector and return the result.

        Parameters
        ----------
        displacement: 2D array of angles
            The displacement by which to move the annotation.

        Returns
        -------
        `~imephu.annotation.general.CircleAnnotation`
            The annotation resulting from the translation.
        """
        translated_annotation = deepcopy(self)
        translated_center = translate(self._center, displacement)
        translated_annotation._center = translated_center
        return translated_annotation



from imephu.annotation.general.circle import CircleAnnotation
from imephu.annotation.general.empty import EmptyAnnotation
from imephu.annotation.general.group import GroupAnnotation
from imephu.annotation.general.line_path import LinePathAnnotation
from imephu.annotation.general.rectangle import RectangleAnnotation
from imephu.annotation.general.text import TextAnnotation

from imephu.annotation.general.arrow import ArrowAnnotation  # isort:skip
from imephu.annotation.general.crosshairs import CrosshairsAnnotation  # isort: skip

__all__ = [
    "ArrowAnnotation",
    "CircleAnnotation",
    "CrosshairsAnnotation",
    "EmptyAnnotation",
    "GroupAnnotation",
    "LinePathAnnotation",
    "RectangleAnnotation",
    "TextAnnotation",
]



from astropy import units as u
from astropy.coordinates import SkyCoord
from astropy.wcs import WCS
from imephu.annotation.general import CircleAnnotation, GroupAnnotation, TextAnnotation
from imephu.geometry import translate


def field_of_view_annotation(fits_center: SkyCoord, wcs: WCS) -> GroupAnnotation:
    """
    Return an annotation with the Salticam field of view.

    Parameters
    ----------
    fits_center: `~astropy.coordinates.SkyCoord`
        The central position of the finder chart, in right ascension and declination.
    wcs: `~astropy.wcs.WCS`
        WCS object.

    Returns
    -------
    `~imephu.annotation.general.GroupAnnotation`
        An annotation with the Salticam field of view.
    """
    fov_annotation = CircleAnnotation(
        fits_center, 5 * u.arcmin, wcs=wcs, edgecolor="green"
    )
    label_position = translate(fits_center, (-3.6, 3.6) * u.arcmin)
    name_annotation = TextAnnotation(
        label_position,
        "SCAM",
        wcs=wcs,
        style="italic",
        weight="bold",
        size="large",
        horizontalalignment="left",
        color=(0, 0, 1),
    )
    return GroupAnnotation([fov_annotation, name_annotation])



from copy import deepcopy
from typing import Any, Sequence, Tuple, Union

from astropy.coordinates import Angle, SkyCoord
from astropy.visualization.wcsaxes import WCSAxes
from astropy.wcs import WCS
from matplotlib.patches import PathPatch
from matplotlib.path import Path

from imephu.annotation import Annotation
from imephu.geometry import rotate, sky_position_to_pixel, translate


class LinePathAnnotation(Annotation):
    """An annotation consisting of straight lines between vertices.

    The path displayed by this annotation is defined by a set of vertices, which are
    connected by straight lines on the finder chart. (The connecting lines in general
    will refer to curved lines on the sky.)

    By default, the path is assumed to be closed, i.e. the last vertex is connected to
    the first by a line. This behavior may be changed with the ``closed`` parameter.

    Parameters
    ----------
    vertices: sequence of `~astropy.coordinates.SkyCoord`
        The path vertices, as an array or sequence of sky coordinates (with right
        ascension and declination).
    wcs: `~astropy.wcs.WCS`
        WCS object.
    closed: bool, default: True
        Whether the path is closed, i.e. whether the last vertex is connected to the
        first by a line.
    edgecolor: color, default: "black"
        The edge color.
    facecolor: color, default: "none"
        The filling color.
    **kwargs: dict, optional
        Additional keyword arguments, which will be passed to Matplotlib's
        `~matplotlib.patches.PathPatch` patch constructor.
    """

    def __init__(
        self,
        vertices: Sequence[SkyCoord],
        wcs: WCS,
        closed: bool = True,
        edgecolor: Union[str, Tuple[float, float, float]] = "black",
        facecolor: Union[str, Tuple[float, float, float]] = "none",
        **kwargs: Any,
    ):
        if closed:
            # Matplotlib will ignore the last vertex for a closed path; we avoid this by
            # having that vertex twice
            self._vertices = list(vertices)[:] + [vertices[-1]]
        else:
            self._vertices = list(vertices)
        self._wcs = wcs
        self._closed = closed
        self._kwargs = deepcopy(kwargs)
        self._kwargs["edgecolor"] = edgecolor
        self._kwargs["facecolor"] = facecolor

    def add_to(self, ax: WCSAxes) -> None:
        """Add the line path to a finder chart.

        Parameters
        ----------
        ax: `~astropy.visualization.wcsaxes.WCSAxes`
            Plot axes.
        """
        vertices_px = [
            sky_position_to_pixel(vertex, self._wcs) for vertex in self._vertices
        ]
        path = Path(vertices_px, closed=self._closed)  # noqa
        path_patch = PathPatch(path, **self._kwargs)
        ax.add_patch(path_patch)

    def rotate(self, pivot: SkyCoord, angle: Angle) -> "LinePathAnnotation":
        """Rotate this annotation around a pivot and return the result.

        The rotation angle is an angle on the sky, measured from north to east.

        Parameters
        ----------
        pivot: `~astropy.coordinates.SkyCoord`
            Point around which to rotate the annotation.
        angle: `~astropy.coordinates.Angle`
            Angle of rotation, measured from north towards the east.

        Returns
        -------
        `~imephu.annotation.general.LinePathAnnotation`
            The annotation resulting from the rotation.
        """
        rotated_annotation = deepcopy(self)
        rotated_vertices = [
            rotate(vertex, pivot, angle, self._wcs) for vertex in self._vertices
        ]
        rotated_annotation._vertices = rotated_vertices
        return rotated_annotation

    def translate(self, displacement: Angle) -> "LinePathAnnotation":
        """
        Move this annotation along a displacement vector and return the result.

        Parameters
        ----------
        displacement: 2D array of angles
            The displacement by which to move the annotation.

        Returns
        -------
        `~imephu.annotation.general.LinePathAnnotation`
            The annotation resulting from the translation.
        """
        translated_annotation = deepcopy(self)
        translated_vertices = [
            translate(vertex, displacement) for vertex in self._vertices
        ]
        translated_annotation._vertices = translated_vertices
        return translated_annotation



from copy import deepcopy

from astropy.coordinates import Angle, SkyCoord
from astropy.visualization.wcsaxes import WCSAxes

from imephu.annotation import Annotation


class EmptyAnnotation(Annotation):
    """An annotation for adding nothing to a finder chart.

    In other words, this annotation just leaves the finder chart as is.

    This is a convenience class for cases where you formally need to pass an annotation,
    but there is nothing to annotate with.
    """

    def add_to(self, ax: WCSAxes) -> None:
        """Add nothing to a finder chart, leaving it as is.

        ax: `~astropy.visualization.wcsaxes.WCSAxes`
            WCS axes object.
        """
        pass

    def rotate(self, pivot: SkyCoord, angle: Angle) -> "Annotation":
        """Return a copy of this annotation.

        Parameters
        ----------
        pivot: `~astropy.coordinates.SkyCoord`
            Point around which to rotate the annotation.
        angle: `~astropy.coordinates.Angle`
            Angle of rotation, measured from north to east.

        Returns
        -------
        `~imephu.annotation.general.CircleAnnotation`
            A copy of this annotation.
        """
        return deepcopy(self)

    def translate(self, displacement: Angle) -> "Annotation":
        """
        Return a copy of this annotation.

        Parameters
        ----------
        displacement: 2D array of angles
            The displacement by which to move the annotation.

        Returns
        -------
        `~imephu.annotation.general.CircleAnnotation`
            A copy of this annotation.
        """
        return deepcopy(self)



import math
from math import atan2, cos, sin
from typing import Any, Tuple, cast

import numpy as np
import numpy.typing as npt
from astropy import units as u
from astropy.coordinates import Angle, SkyCoord
from astropy.wcs import WCS
from astropy.wcs.utils import pixel_to_skycoord, skycoord_to_pixel


def rotate(v: SkyCoord, pivot: SkyCoord, angle: Angle, wcs: WCS) -> SkyCoord:
    """Rotate a point around a pivot.

    Both the pivot and the point to rotate are assumed to be given as right ascension
    and declination. The rotation angle is taken to be the angle on the sky. If the
    pixel scales for right ascension and declination differ, the angle seen on the
    finder chart will differ from this angle.

    The angle is measured from north to south. Hence it depends on the orientation of
    the coordinate axes whether a positive angle corresponds to a clockwise or
    anti-clockwise orientation.

    Parameters
    ----------
    v: `~astropy.coordinates.SkyCoord`
        The point to rotate.
    pivot: `~astropy.coordinates.SkyCoord`
        The point around which the point ``v`` is rotated.
    angle: `~astropy.coordinates.Angle`
        The angle of rotation, measured from north to east.
    wcs: `~astropy.wcs.WCS`
        WCS object.

    Returns
    -------
    `~astropy.coordinates.SkyCoord`
        The point after the rotation.
    """
    # Convert from world coordinates to pixels
    v_px = sky_position_to_pixel(v, wcs)
    pivot_px = sky_position_to_pixel(pivot, wcs)

    # Correct the angle for different pixel scales
    sx, sy = pixel_scales(wcs)
    angle_rad = angle.to_value(u.rad)
    x_angle_sky = cos(angle_rad)
    y_angle_sky = sin(angle_rad)
    x_angle_px = x_angle_sky
    y_angle_px = float(sx / sy) * y_angle_sky
    angle_px = atan2(y_angle_px, x_angle_px)

    # If necessary, change the sign of the angle to accommodate the orientation of the
    # axes on the finder chart.
    if not _is_positive_angle_anti_clockwise(wcs):
        angle_px *= -1

    # Perform the rotation
    rotation: Any = np.array(
        [[cos(angle_px), -sin(angle_px)], [sin(angle_px), cos(angle_px)]]
    )
    diff_vector_px: Any = np.array(v_px) - np.array(pivot_px)
    rotated_v_px = np.array(pivot_px) + rotation @ diff_vector_px

    # Convert back from pixels to world coordinates
    return pixel_to_sky_position(rotated_v_px, wcs)


def translate(v: SkyCoord, displacement: Angle) -> SkyCoord:
    """Move a point by a displacement vector.

    The point to move is assumed to be given as right ascension and declination. The
    displacement vector is assumed to be a vector on the sky, with real angles, as
    described for the `~sky_vector_to_pixel` method.

    Parameters
    ----------
    v: `~astropy.coordinates.SkyCoord`
        Point to move.
    displacement: 2D array of angles
        Vector by which to move the point.

    Returns
    -------
    `~astropy.coordinates.SkyCoord`
        The point after moving.
    """
    # Convert from sky coordinates to real angles
    displacement_delta_ra = displacement[0] / cos(v.dec.to_value(u.rad))
    displacement_delta_dec = displacement[1]

    # Perform the translation
    translated_v_ra = v.ra + displacement_delta_ra
    translated_v_dec = v.dec + displacement_delta_dec
    return SkyCoord(ra=translated_v_ra, dec=translated_v_dec)


def pixel_scales(wcs: WCS) -> Tuple[Angle, Angle]:
    """
    Calculate the pixel scales of a WCS object.

    For a WCS returns pixel scales along each axis of the image pixel at
    the ``CRPIX`` location once it is projected onto the
    "plane of intermediate world coordinates" as defined in
    `Greisen & Calabretta 2002, A&A, 395, 1061
    <https://ui.adsabs.harvard.edu/abs/2002A%26A...395.1061G>`_.

    .. note::
        This function is concerned **only** about the transformation
        "image plane"->"projection plane" and **not** about the
        transformation "celestial sphere"->"projection plane"->"image plane".
        Therefore, this function ignores distortions arising due to
        non-linear nature of most projections.

    .. note::
        In order to compute the scales corresponding to celestial axes only,
        make sure that the input `~astropy.wcs.WCS` object contains
        celestial axes only, e.g., by passing in the
        `~astropy.wcs.WCS.celestial` WCS object.

    The code has been adapted from APLpy.

    Parameters
    ----------
    wcs : `~astropy.wcs.WCS`
        A world coordinate system object.

    Returns
    -------
    scale : tuple of `~astropy.coordinates.Angle`
        A vector (`~numpy.ndarray`) of projection plane increments
        corresponding to each pixel side (axis). The units of the returned
        results are the same as the units of `~astropy.wcs.Wcsprm.cdelt`,
        `~astropy.wcs.Wcsprm.crval`, and `~astropy.wcs.Wcsprm.cd` for
        the celestial WCS and can be obtained by inquiring the value
        of `~astropy.wcs.Wcsprm.cunit` property of the input
        `~astropy.wcs.WCS` WCS object.

    See Also
    --------
    astropy.wcs.utils.proj_plane_pixel_area

    """
    scale_factors = np.sqrt((wcs.pixel_scale_matrix**2).sum(axis=0, dtype=float))
    return (
        Angle(scale_factors[0], wcs.world_axis_units[0]),
        Angle(scale_factors[1], wcs.world_axis_units[1]),
    )


def sky_position_to_pixel(position: SkyCoord, wcs: WCS) -> npt.NDArray[np.float_]:
    """Convert a sky position to the corresponding pixel coordinates.

    Parameters
    ----------
    position: `~astropy.coordinates.SkyCoord`
        Sky position as right ascension and declination.
    wcs: `~astropy.wcs.WCS`
        WCS object.

    Returns
    -------
    `~numpy.ndarray` of `float`
        The pixel coordinates.
    """
    pixel_coords = skycoord_to_pixel(coords=position, wcs=wcs)
    return np.array([float(pixel_coords[0].item()), float(pixel_coords[1].item())])


def pixel_to_sky_position(position_px: npt.NDArray[np.float_], wcs: WCS) -> SkyCoord:
    """Convert pixel coordinates to the corresponding sky position.

    Parameters
    ----------
    position_px: `~numpy.ndarray` of `float`
        Pixel coordinates.
    wcs: `~astropy.wcs.WCS`
        WCS object.

    Returns
    -------
    `~astropy.coordinates.SkyCoord`
        The sky coordinates.
    """
    return pixel_to_skycoord(position_px[0], position_px[1], wcs=wcs)  # noqa


def _is_positive_angle_anti_clockwise(wcs: WCS) -> bool:
    """
    Return whether for a given WCS positive angles are anti-clockwise when plotted.

    Angles on the sky are measured from north to east, and depending on the orientation
    of the axes on the finder chart this may correspond to a mathematically positive
    ("anti-clockwise") or negative ("clockwise") angle.

    Parameters
    ----------
    wcs: `~astropy.wcs.WCS`
        WCS object.

    Returns
    -------
    bool
        ``True`` if angles from north to east correspond to mathematically positive
        ("anti-clockwise") angles on a finder chart, ``False`` otherwise.
    """
    origin_px: npt.NDArray[np.float_] = np.array((0.0, 0.0))
    origin_on_sky = pixel_to_sky_position(origin_px, wcs)
    origin_ra = origin_on_sky.ra
    origin_dec = origin_on_sky.dec

    # A "point to the north" on a finder chart
    point_to_north_ra = origin_ra
    point_to_north_dec = origin_dec + 1 * u.arcmin
    point_to_north = SkyCoord(ra=point_to_north_ra, dec=point_to_north_dec)
    point_to_north_px = sky_position_to_pixel(point_to_north, wcs)

    # A "point to the north-east" on a finder chart
    point_to_north_east_ra = origin_ra + 1 * u.arcmin
    point_to_north_east_dec = origin_dec + 1 * u.arcmin
    point_to_north_east = SkyCoord(
        ra=point_to_north_east_ra, dec=point_to_north_east_dec
    )
    point_to_north_east_px = sky_position_to_pixel(point_to_north_east, wcs)

    # A "point to the north-east" on a finder chart
    point_to_north_west_ra = origin_ra - 1 * u.arcmin
    point_to_north_west_dec = origin_dec + 1 * u.arcmin
    point_to_north_west = SkyCoord(
        ra=point_to_north_west_ra, dec=point_to_north_west_dec
    )
    point_to_north_west_px = sky_position_to_pixel(point_to_north_west, wcs)

    # The difference between the point to the north and the origin gives a vector
    # pointing north - but as the origin is (0, 0) this is just the point to the north.
    # The same argument applies for the vector pointing north-east.
    north_vector_px = point_to_north_px
    north_east_vector_px = point_to_north_east_px
    north_west_vector_px = point_to_north_west_px

    # Get the angle between the vectors to the north and to the north-east. Due to
    # symmetry, this is the same as the angle between the vectors to the north and to
    # north-west
    cos_angle = np.dot(north_vector_px, north_east_vector_px) / (
        np.linalg.norm(north_vector_px) * np.linalg.norm(north_east_vector_px)
    )
    angle = math.acos(cos_angle)

    # Rotate the vector to the north (in anti-clockwise direction) by this angle
    rotation: npt.NDArray[np.float_] = np.array(
        [[cos(angle), -sin(angle)], [sin(angle), cos(angle)]]
    )
    rotated_north_vector_px = np.dot(rotation, north_vector_px)

    # If a rotation from north to east corresponds to an anti-clockwise rotation on a
    # finder chart, the rotated vector must aligned with the vector to the north-east.
    # This implies that their scalar product then must be larger than that between the
    # rotated vector and the vector to the north-west.
    return cast(
        bool,
        np.dot(rotated_north_vector_px, north_east_vector_px)
        > np.dot(rotated_north_vector_px, north_west_vector_px),
    )



from copy import deepcopy
from typing import Any, Literal, Sequence, Tuple, Union

import numpy as np
import numpy.typing as npt
from astropy.coordinates import Angle, SkyCoord
from astropy.visualization.wcsaxes import WCSAxes
from astropy.wcs import WCS

from imephu.annotation import Annotation
from imephu.geometry import rotate, sky_position_to_pixel, translate


class TextAnnotation(Annotation):
    """An annotation for adding text to a plot.

    The text position can be specified as a position on the sky (as an instance of
    `~astropy.coordinates.SkyCoord`) or as a point in axes coordinates (as an array of
    `float` values. Axes coordinates range from (0, 0) (bottom left of the axes) to
    (1, 1) (top right of the axes).

    You can only rotate and translate the annotation if the text position is specified
    as a position on the sky.

    .. note::
        The `rotate` method rotates the text position, but not the text itself. If you
        want to rotate the text rather than the text position, you should use the
        ``rotation`` parameter with the angle in degrees.


    Parameters
    ----------
    position: `~astropy.coordinates.SkyCoord` or list of float
        The right ascension and declination of the point where to put the text.
    s: str
        The text.
    wcs: `~astropy.wcs.WCS`
        WCS object.
    color: color, default: "black"
        The text color.
    horizontalalignment: {"center", "left", "right"}, default: "center"
        The horizontal alignment of the text.
    verticalalignment: {"baseline", "bottom", "center", "center_baseline", "top"}, default: "center"
        The vertical alignment of the text.
    clip_on: bool, default: True
        Whether to clip the text at the plot boundaries.
    **kwargs: dict, optional
        Additional keyword arguments, which will be passed to Matplotlib's
        `~matplotlib.axes.Axes.text` method for adding text to plot Axes.
    """  # noqa: E501

    def __init__(
        self,
        position: Union[SkyCoord, Sequence[float], npt.NDArray[np.float_]],
        s: str,
        wcs: WCS,
        color: Union[str, Tuple[float, float, float]] = "black",
        horizontalalignment: Literal["center", "left", "right"] = "center",
        verticalalignment: Literal[
            "baseline", "bottom", "center", "center_baseline", "top"
        ] = "center",
        clip_on: bool = True,
        **kwargs: Any,
    ):
        self._position = position
        self._s = s
        self._wcs = wcs
        self._kwargs = deepcopy(kwargs)
        self._kwargs["color"] = color
        self._kwargs["horizontalalignment"] = horizontalalignment
        self._kwargs["verticalalignment"] = verticalalignment
        self._kwargs["clip_on"] = clip_on

    def add_to(self, ax: WCSAxes) -> None:
        """Add the text to a finder chart.

        ax:  `~astropy.visualization.wcsaxes.WCSAxes`
            WCS axes object.
        """
        if type(self._position) == SkyCoord:
            position_px = sky_position_to_pixel(self._position, self._wcs)
            ax.text(position_px[0], position_px[1], self._s, **self._kwargs)  # noqa
        else:
            ax.text(
                self._position[0],
                self._position[1],
                self._s,
                transform=ax.transAxes,
                **self._kwargs,
            )

    def rotate(self, pivot: SkyCoord, angle: Angle) -> "TextAnnotation":
        """Rotate this annotation around a pivot and return the result.

        The rotation angle is an angle on the sky, measured from north to east.

        Parameters
        ----------
        pivot: `~astropy.coordinates.SkyCoord`
            Point around which to rotate the annotation.
        angle: `~astropy.coordinates.Angle`
            Angle of rotation, measured from north to east.

        Returns
        -------
        `~imephu.annotation.general.TextAnnotation`
            The annotation resulting from the rotation.
        """
        if type(self._position) is not SkyCoord:
            raise NotImplementedError(
                "The text can only be rotated if its position is an instance of "
                "astropy.coord.SkyCoord."
            )
        rotated_annotation = deepcopy(self)
        rotated_position = rotate(self._position, pivot, angle, self._wcs)
        rotated_annotation._position = rotated_position
        return rotated_annotation

    def translate(self, displacement: Angle) -> "TextAnnotation":
        """
        Move this annotation along a displacement vector and return the result.

        Parameters
        ----------
        displacement: 2D array of angles
            The displacement by which to move the annotation.

        Returns
        -------
        `~imephu.annotation.general.TextAnnotation`
            The annotation resulting from the translation.
        """
        if type(self._position) is not SkyCoord:
            raise NotImplementedError(
                "The text can only be translated if its position is an instance of "
                "astropy.coord.SkyCoord."
            )
        translated_annotation = deepcopy(self)
        translated_position = translate(self._position, displacement)
        translated_annotation._position = translated_position
        return translated_annotation



from typing import Any, Tuple, Union

from astropy import units as u
from astropy.coordinates import Angle, SkyCoord
from astropy.wcs import WCS

from imephu.annotation.general import GroupAnnotation, LinePathAnnotation
from imephu.geometry import translate


class CrosshairsAnnotation(GroupAnnotation):
    """An annotation for plotting crosshairs.

    The crosshairs are defined by their center (i.e. where the two lines intersect), the
    size (i.e. the length of the lines) and the angle of rotation around the center.

    The lines bisect each other. The crosshair lines are in the direction of right
    ascension and declination.

    Parameters
    ----------
    center: `~astropy.coordinates.SkyCoord`
        The center of the crosshairs, i.e. the point of intersection of the two lines,
        as a position on the sky (in right ascension and declination).
    size: `~astropy.coordinates.Angle`
        The length of each line, as an angle on the sky.
    wcs: `~astropy.wcs.WCS`
        WCS object.
    color: color, default: "black"
        The line color.
    **kwargs: dict, optional
        Additional keyword arguments, which will be passed to Matplotlib's
        `~matplotlib.patches.PathPatch` patch constructor.
    """

    def __init__(
        self,
        center: SkyCoord,
        size: Angle,
        wcs: WCS,
        color: Union[str, Tuple[float, float, float]] = "black",
        **kwargs: Any,
    ) -> None:
        super().__init__(
            items=CrosshairsAnnotation._crosshair_lines(
                center=center, size=size, wcs=wcs, color=color, **kwargs
            )
        )

    @staticmethod
    def _crosshair_lines(
        center: SkyCoord,
        size: Angle,
        wcs: WCS,
        color: Union[str, Tuple[float, float, float]] = "black",
        **kwargs: Any,
    ) -> Tuple[LinePathAnnotation, LinePathAnnotation]:
        size_arcsec = size.to_value(u.arcsec)
        line_ra = LinePathAnnotation(
            vertices=[
                translate(center, Angle((-size_arcsec / 2, 0) * u.arcsec)),
                translate(center, Angle((size_arcsec / 2, 0) * u.arcsec)),
            ],
            wcs=wcs,
            edgecolor=color,
            **kwargs,
        )
        line_dec = LinePathAnnotation(
            vertices=[
                translate(center, Angle((0, -size_arcsec / 2) * u.arcsec)),
                translate(center, Angle((0, size_arcsec / 2) * u.arcsec)),
            ],
            wcs=wcs,
            edgecolor=color,
            **kwargs,
        )
        return line_ra, line_dec



"""
imephu is a package for generating astronomical finder charts.

The finder charts are primarily intended for the Southern African Large Telescope
(SALT).
"""

from importlib.metadata import version

__version__ = version("imephu")



from typing import Any, List, Tuple, Union, cast

import numpy as np
from astropy import units as u
from astropy.coordinates import Angle, SkyCoord
from astropy.wcs import WCS

from imephu.annotation.general.line_path import LinePathAnnotation
from imephu.geometry import translate


class RectangleAnnotation(LinePathAnnotation):
    """An annotation for plotting a rectangle on a finder chart.

    The rectangle is defined by its center, width and height. The center must be a sky
    position in right ascension and declination, the width and height must be an angle
    on the sky. The width is taken to be the extension in the right ascension direction,
    the height the extension in the declination direction.

    The corners of the rectangle are mapped to their corresponding pixel positions on
    the finder chart. This implies that in general the figure plotted on the finder
    chart will be a quadrilateral, but not exactly a rectangle.

    .. note::
        The center rather than "bottom left" corner is specified, as this is more
        convenient in the context of finder charts.

    Parameters
    ----------
    center: `~astropy.coordinates.SkyCoord`
        The center of the rectangle, in right ascension and declination.
    width: `~astropy.coordinates.Angle`
        The width of the rectangle as an angle on the sky in the direction of the
        right ascension.
    height: `~astropy.coordinates.Angle`
        The height of the rectangle as an angle on the sky in the direction of the
        declination.
    wcs: `~astropy.wcs.WCS`
        WCS object.
    edgecolor: color, default: "black"
        The edge color.
    facecolor: color, default: "none"
        The filling color.
    **kwargs: dict, optional
        Additional keyword arguments, which will be passed to Matplotlib's
        `~matplotlib.patches.PathPatch` patch constructor.
    """

    def __init__(
        self,
        center: SkyCoord,
        width: Angle,
        height: Angle,
        wcs: WCS,
        edgecolor: Union[str, Tuple[float, float, float]] = "black",
        facecolor: Union[str, Tuple[float, float, float]] = "none",
        **kwargs: Any,
    ):
        super().__init__(
            vertices=RectangleAnnotation._corners(center, width, height),
            wcs=wcs,
            closed=True,
            edgecolor=edgecolor,
            facecolor=facecolor,
            **kwargs,
        )

    @staticmethod
    def _corners(center: SkyCoord, width: Angle, height: Angle) -> List[SkyCoord]:
        corners: List[SkyCoord] = []
        half_width_arcsec = width.to_value(u.arcsec) / 2
        half_height_arcsec = height.to_value(u.arcsec) / 2
        directions = [(-1, -1), (1, -1), (1, 1), (-1, 1)]
        for direction in directions:
            center_to_corner = (
                np.array(
                    (
                        direction[0] * half_width_arcsec,
                        direction[1] * half_height_arcsec,
                    )
                )
                * u.arcsec
            )
            corner = translate(center, center_to_corner)
            corners.append(corner)
        return corners

    def rotate(self, pivot: SkyCoord, angle: Angle) -> "RectangleAnnotation":
        """Rotate this annotation around a pivot and return the result.

        The rotation angle is an angle on the sky, measured from north to east.

        Parameters
        ----------
        pivot: `~astropy.coordinates.SkyCoord`
            Point around which to rotate the annotation.
        angle: `~astropy.coordinates.Angle`
            Angle of rotation, measured from north towards the east.

        Returns
        -------
        `~imephu.annotation.general.RectangleAnnotation`
            The annotation resulting from the rotation.
        """
        return cast(RectangleAnnotation, super().rotate(pivot, angle))

    def translate(self, displacement: Angle) -> "RectangleAnnotation":
        """
        Move this annotation along a displacement vector and return the result.

        Parameters
        ----------
        displacement: 2D array of angles
            The displacement by which to move the annotation.

        Returns
        -------
        `~imephu.annotation.general.RectangleAnnotation`
            The annotation resulting from the translation.
        """
        return cast(RectangleAnnotation, super().translate(displacement))



from __future__ import annotations

import dataclasses
import os
from dataclasses import dataclass
from datetime import datetime
from typing import Any, BinaryIO, Generator, List, Optional, Tuple, Union

from astropy import units as u
from astropy.coordinates import Angle, SkyCoord
from astropy.wcs import WCS
from imephu.annotation.general import GroupAnnotation
from imephu.annotation.motion import motion_annotation
from imephu.finder_chart import FinderChart
from imephu.salt.annotation import nir, rss, telescope
from imephu.salt.utils import MosMask
from imephu.utils import (
    Ephemeris,
    MagnitudeRange,
    ephemerides_magnitude_range,
    mid_position,
)

_FINDER_CHART_SIZE = 10 * u.arcmin


class SaltFinderChart(FinderChart):
    """A finder chart for an astronomical observation.

    The finder chart is generated from a FITS file. By default the finder chart just
    shows the region with an overlaid coordinate grid. The axes show WCS coordinates.
    Use the `add_annotation` method to add more content to the finder chart.

    By default, the size of the finder chart is the same as that of the FITS file. You
    may set the `max_size` property to choose a non-default maximum size.

    You can display the finder chart on the screen or save it as a file.

    This class uses Matplotlib for generating the finder chart.

    .. warning:: Be careful when working with several finder charts as all the finder
       charts use the same Matplotlib figure.

    Parameters
    ----------
    name: `str`, `path-like` or `binary file-like`
        FITS file to display.

    target: `imephu.salt.finder_chart.Target`
        Target for the finder chart.

    max_size: `~astropy.coordinates.Angle`
        Maximum size of the generated finder chart.
    """

    def __init__(
        self,
        name: Union[str, BinaryIO, os.PathLike[Any]],
        target: Target,
        max_size: Optional[Angle] = None,
    ):
        super().__init__(name, max_size)

        right_ascension_deg = target.position.ra.to(u.deg).value
        declination_deg = target.position.dec.to(u.deg).value
        self.add_metadata("right_ascension", f"{right_ascension_deg} deg")
        self.add_metadata("declination", f"{declination_deg} deg")


@dataclass
class Target:
    """Target properties.

    Attributes
    ----------
    name: `str`
        The target name.
    position: `~astropy.coordinates.SkyCoord`
        The target position, as a right ascension and declination. This is taken to be
        the center of the finder chart.
    magnitude_range: `imephu.utils.MagnitudeRange`, optional
        The magnitude range of the target.
    """

    name: str
    position: SkyCoord
    magnitude_range: Optional[MagnitudeRange]


@dataclass
class GeneralProperties:
    """Properties which are not specific to a particular instrument.

    Attributes
    ----------
    target: `Target`
        The target to find using this finder chart. Its position is taken to be the
        center of the finder chart.
    position_angle: `~astropy.coordinates.Angle`
        The position angle, as angle on the sky measured from north to east.
    automated_position_angle: `bool`
        Whether the position angle has been calculated automatically.
    proposal_code: `str`
        The proposal code.
    pi_family_name: `str`
        The family name of the Principal Investigator.
    survey: `str`, optional
        The image survey from which the FITS image is taken.
    """

    target: Target
    position_angle: Angle
    automated_position_angle: bool
    proposal_code: str
    pi_family_name: str
    survey: Optional[str] = None


def salticam_finder_chart(
    fits: Union[str, BinaryIO, os.PathLike[Any]],
    general: GeneralProperties,
    is_slot_mode: bool = False,
) -> FinderChart:
    """Return the finder chart for a Salticam observation.

    Parameters
    ----------
    fits: `str`, `path-like` or `binary file-like`
        FITS file to display.
    general: `GeneralProperties`
        Properties which are not specific to the instrument.
    is_slot_mode: `bool`, default: False
        Whether the observation is a slot mode one.

    Returns
    -------
    `~imephu.finder_chart.FinderChart`
        The finder chart for a Salticam observation.
    """
    finder_chart = SaltFinderChart(fits, general.target)
    annotation = _salticam_observation_annotation(
        general=general, is_slot_mode=is_slot_mode, wcs=finder_chart.wcs
    )
    finder_chart.add_annotation(annotation)
    return finder_chart


def rss_imaging_finder_chart(
    fits: Union[str, BinaryIO, os.PathLike[Any]],
    general: GeneralProperties,
    is_slot_mode: bool = False,
) -> FinderChart:
    """Return the finder chart for an RSS imaging observation.

    Parameters
    ----------
    fits: `str`, `path-like` or `binary file-like`
        FITS file to display.
    general: `GeneralProperties`
        Properties which are not specific to the instrument.
    is_slot_mode: `bool`, default: False
        Whether the observation is a slot mode one.

    Returns
    -------
    `~imephu.finder_chart.FinderChart`
        The finder chart for an RSS imaging observation.
    """
    finder_chart = SaltFinderChart(fits, general.target)
    annotation = _rss_imaging_observation_annotation(
        general=general, is_slot_mode=is_slot_mode, wcs=finder_chart.wcs
    )
    finder_chart.add_annotation(annotation)
    return finder_chart


def rss_longslit_finder_chart(
    fits: Union[str, BinaryIO, os.PathLike[Any]],
    general: GeneralProperties,
    slit_width: Angle,
    slit_height: Angle,
    reference_star: Optional[SkyCoord],
) -> FinderChart:
    """Return the finder chart for an RSS longslit observation.

    Parameters
    ----------
    fits: `str`, `path-like` or `binary file-like`
        FITS file to display.
    general: `GeneralProperties`
        Properties which are not specific to the instrument.
    slit_width: `~astropy.coordinates.Angle`
        The slit width, as an angle on the sky.
    slit_height: `~astropy.coordinates.Angle`
        The slit height, as an angle on the sky.
    reference_star: `~astropy.coordinates.SkyCoord`, optional
        The position of a reference star, in right ascension and declination.

    Returns
    -------
    `~imephu.finder_chart.FinderChart`
        The finder chart for an RSS longslit observation.
    """
    finder_chart = SaltFinderChart(fits, general.target)
    annotation = _rss_longslit_observation_annotation(
        general=general,
        slit_width=slit_width,
        slit_height=slit_height,
        reference_star=reference_star,
        wcs=finder_chart.wcs,
    )
    finder_chart.add_annotation(annotation)
    return finder_chart


DEFAULT_REF_STAR_BOX_SIZE = Angle(5 * u.arcsec)


def rss_mos_finder_chart(
    fits: Union[str, BinaryIO, os.PathLike[Any]],
    general: GeneralProperties,
    mos_mask: MosMask,
    reference_star_box_width: Angle = DEFAULT_REF_STAR_BOX_SIZE,
) -> FinderChart:
    """Return the finder chart for an RSS MOS observation.

    Parameters
    ----------
    fits: `str`, `path-like` or `binary file-like`
        FITS file to display.
    general: `GeneralProperties`
        Properties which are not specific to the instrument.
    mos_mask: `~imephu.salt.utils.MosMask`
        The MOS mask.
    reference_star_box_width: `~astropy.coordinates.Angle`, default: 5 arcseconds
        The width (and height) of the boxes around reference stars, as an angle on the
        sky.

    Returns
    -------
    `~imephu.finder_chart.FinderChart`
        The finder chart for an RSS MOS observation.
    """
    finder_chart = SaltFinderChart(fits, general.target)
    annotation = _rss_mos_observation_annotation(
        general=general,
        mos_mask=mos_mask,
        reference_star_box_width=reference_star_box_width,
        wcs=finder_chart.wcs,
    )
    finder_chart.add_annotation(annotation)
    return finder_chart


def rss_fabry_perot_finder_chart(
    fits: Union[str, BinaryIO, os.PathLike[Any]],
    general: GeneralProperties,
) -> FinderChart:
    """Return the finder chart for an RSS Fabry-Pérot observation.

    Parameters
    ----------
    fits: `str`, `path-like` or `binary file-like`
        FITS file to display.
    general: `GeneralProperties`
        Properties which are not specific to the instrument.

    Returns
    -------
    `~imephu.finder_chart.FinderChart`
        The finder chart for an RSS Fabry-Pérot observation.
    """
    finder_chart = SaltFinderChart(fits, general.target)
    annotation = _rss_fabry_perot_observation_annotation(
        general=general, wcs=finder_chart.wcs
    )
    finder_chart.add_annotation(annotation)
    return finder_chart


def nir_finder_chart(
    fits: Union[str, BinaryIO, os.PathLike[Any]],
    general: GeneralProperties,
    reference_star: Optional[SkyCoord],
    bundle_separation: Angle,
) -> FinderChart:
    """Return the finder chart for an NIR observation.

    If coordinates for a reference star are passed, they define the center of the finder
    chart, and the sky bundle is centered on the target coordinates.

    Parameters
    ----------
    fits: `str`, `path-like` or `binary file-like`
        FITS file to display.
    reference_star: `~astropy.coordinates.SkyCoord`, optional
        The position of the reference star, as a position on the sky, in right ascension
        and declination. This is taken to be the center of the finder chart.
    bundle_separation: ~astropy.coordinates.Angle`
        The separation between the science fiber bundle and the sky fiber bundles, as an
        angle on the sky. The separation is measured between the center of the science
        bundle and the midpoint of the line between the centers of the sky bundles.

    Returns
    -------
    `~imephu.finder_chart.FinderChart`
        The finder chart for an NIR observation.
    """
    finder_chart = SaltFinderChart(fits, general.target)
    annotation = _nir_observation_annotation(
        general=general,
        reference_star=reference_star,
        bundle_separation=bundle_separation,
        wcs=finder_chart.wcs,
    )
    finder_chart.add_annotation(annotation)
    return finder_chart


def hrs_finder_chart(
    fits: Union[str, BinaryIO, os.PathLike[Any]], general: GeneralProperties
) -> FinderChart:
    """Return the finder chart for an HRS observation.

    Parameters
    ----------
    fits: `str`, `path-like` or `binary file-like`
        FITS file to display.
    general: `GeneralProperties`
        Properties which are not specific to the instrument.

    Returns
    -------
    `~imephu.finder_chart.FinderChart`
        The finder chart for an HRS observation.
    """
    finder_chart = SaltFinderChart(fits, general.target)
    annotation = _hrs_observation_annotation(general=general, wcs=finder_chart.wcs)
    finder_chart.add_annotation(annotation)
    return finder_chart


def moving_target_finder_charts(
    general: GeneralProperties,
    start: datetime,
    end: datetime,
    ephemerides: List[Ephemeris],
    survey: str,
) -> Generator[Tuple[FinderChart, Tuple[datetime, datetime]], None, None]:
    """Return a generator for the finder charts for a moving target in a time interval.

    Parameters
    ----------
    general: `GeneralProperties`
        Properties which are not specific to the instrument.
    start: `~datetime.datetime`
        Start time of the time interval for which finder charts are generated. This must
        be timezone-aware.
    end: `~datetime.datetime`
        End time of the time interval for which finder charts are generated. This must
        be timezone-aware.
    ephemerides: list of `~imephu.utils.Ephemeris`
        The ephemerides for the moving target.
    survey: `str`
        The name of the survey from which the finder chart image should be taken.

    Yields
    ------
    `~imephu.finder_chart.FinderChart`
        The finder chart.
    """

    def _create_finder_chart(ephemerides_: List[Ephemeris]) -> FinderChart:
        midpoint = mid_position(ephemerides_[0].position, ephemerides_[-1].position)
        magnitude_range = ephemerides_magnitude_range(ephemerides_)
        general_ = dataclasses.replace(
            general,
            target=Target(
                name=general.target.name,
                position=midpoint,
                magnitude_range=magnitude_range,
            ),
        )
        finder_chart = FinderChart.from_survey(survey, midpoint, _FINDER_CHART_SIZE)
        annotation = _non_sidereal_annotation(
            general=general_, ephemerides=ephemerides_, wcs=finder_chart.wcs
        )
        finder_chart.add_annotation(annotation)
        return finder_chart

    return FinderChart.for_time_interval(
        start=start,
        end=end,
        ephemerides=ephemerides,
        max_track_length=0.8 * _FINDER_CHART_SIZE,
        create_finder_chart=_create_finder_chart,
    )


def _salticam_observation_annotation(
    general: GeneralProperties, is_slot_mode: bool, wcs: WCS
) -> GroupAnnotation:
    return _imaging_annotation(general, is_slot_mode, wcs)


def _rss_imaging_observation_annotation(
    general: GeneralProperties, is_slot_mode: bool, wcs: WCS
) -> GroupAnnotation:
    return _imaging_annotation(general, is_slot_mode, wcs)


def _rss_longslit_observation_annotation(
    general: GeneralProperties,
    slit_width: Angle,
    slit_height: Angle,
    reference_star: SkyCoord,
    wcs: WCS,
) -> GroupAnnotation:
    observation_annotation = _base_annotations(general, wcs)
    longslit_annotation = rss.longslit_annotation(
        fits_center=general.target.position,
        slit_width=slit_width,
        slit_height=slit_height,
        position_angle=general.position_angle,
        wcs=wcs,
    )
    observation_annotation.add_item(longslit_annotation)
    if reference_star:
        reference_star_annotation = rss.reference_star_annotation(reference_star, wcs)
        observation_annotation.add_item(reference_star_annotation)
    return observation_annotation


def _rss_mos_observation_annotation(
    general: GeneralProperties,
    mos_mask: MosMask,
    reference_star_box_width: Angle,
    wcs: WCS,
) -> GroupAnnotation:
    observation_annotation = _base_annotations(general, wcs)
    mask_annotation = rss.mos_mask_annotation(
        mos_mask=mos_mask,
        wcs=wcs,
        reference_star_box_width=reference_star_box_width,
    )
    observation_annotation.add_item(mask_annotation)
    return observation_annotation


def _rss_fabry_perot_observation_annotation(
    general: GeneralProperties, wcs: WCS
) -> GroupAnnotation:
    return _imaging_annotation(general=general, is_slot_mode=False, wcs=wcs)


def _nir_observation_annotation(
    general: GeneralProperties,
    reference_star: Optional[SkyCoord],
    bundle_separation: Angle,
    wcs: WCS,
) -> GroupAnnotation:
    science_bundle_center = general.target.position
    if reference_star:
        general.target.position = reference_star
    observation_annotation = _base_annotations(general, wcs)
    bundles_annotation = nir.bundles_annotation(
        finder_chart_center=general.target.position,
        science_bundle_center=science_bundle_center,
        bundle_separation=bundle_separation,
        position_angle=general.position_angle,
        wcs=wcs,
    )
    observation_annotation.add_item(bundles_annotation)
    return observation_annotation


def _hrs_observation_annotation(
    general: GeneralProperties, wcs: WCS
) -> GroupAnnotation:
    return _imaging_annotation(general=general, is_slot_mode=False, wcs=wcs)


def _non_sidereal_annotation(
    general: GeneralProperties, ephemerides: List[Ephemeris], wcs: WCS
) -> GroupAnnotation:
    imaging_annotation = _imaging_annotation(
        general=general, is_slot_mode=False, wcs=wcs
    )
    track_annotation = motion_annotation(ephemerides=ephemerides, wcs=wcs)
    return GroupAnnotation(items=[imaging_annotation, track_annotation])


def _imaging_annotation(
    general: GeneralProperties, is_slot_mode: bool, wcs: WCS
) -> GroupAnnotation:
    observation_annotation = _base_annotations(general, wcs)
    magnitude_range = general.target.magnitude_range
    if magnitude_range:
        magnitude_annotation = telescope.magnitude_range_annotation(
            bandpass=magnitude_range.bandpass,
            min_magnitude=magnitude_range.min_magnitude,
            max_magnitude=magnitude_range.max_magnitude,
            wcs=wcs,
        )
        observation_annotation.add_item(magnitude_annotation)
    if is_slot_mode:
        center = general.target.position
        slot_annotation = telescope.slot_annotation(center, general.position_angle, wcs)
        observation_annotation.add_item(slot_annotation)
    return observation_annotation


def _base_annotations(general: GeneralProperties, wcs: WCS) -> GroupAnnotation:
    return telescope.base_annotations(
        target=general.target.name,
        proposal_code=general.proposal_code,
        pi_family_name=general.pi_family_name,
        position_angle=general.position_angle,
        automated_position_angle=general.automated_position_angle,
        survey=general.survey,
        fits_center=general.target.position,
        wcs=wcs,
    )



import dataclasses
import math
from copy import deepcopy
from typing import Any, Tuple

import astropy.units as u
from astropy.coordinates import Angle, SkyCoord
from astropy.visualization.wcsaxes import WCSAxes
from astropy.wcs import WCS
from matplotlib.patches import PathPatch
from matplotlib.path import Path

from imephu.annotation import Annotation
from imephu.geometry import pixel_scales


@dataclasses.dataclass()
class _ScaleBarParameters:
    pixels: float
    angle: float
    units: str


class ScaleBarLineAnnotation(Annotation):
    """An annotation for drawing the line of a scale bar.

    The position of the scale bar's centre must be specified in axes coordinates. Axes
    coordinates range from (0, 0) (bottom left of the axes) to (1, 1) (top right of the
    axes).

    Parameters
    ----------
    left_edge: pair of float
        The position of the left edge of the scale bar's horizontal line, in pixels.
    minimum_length: float
        The minimum length of the scale bar's horizontal line, in pixels.
    wcs: `~astropy.wcs.WCS`
        WCS object.
    color: color, default: "black"
        The color of the scale bar.
    linewidth: float, default: 0.5
        The line width.
    **kwargs: dict, optional
        Additional keyword arguments, which must be valid arguments for both paths and
        text.
    """

    def __init__(
        self,
        left_edge: Tuple[float, float],
        minimum_length: float,
        wcs: WCS,
        color: str = "black",
        linewidth: float = 0.5,
        **kwargs: Any,
    ):
        self._left_edge = left_edge
        self._minimum_length = minimum_length
        self._wcs = wcs
        self._kwargs = deepcopy(kwargs)
        self._kwargs["edgecolor"] = color
        self._kwargs["facecolor"] = color
        self._kwargs["linewidth"] = linewidth

    def add_to(self, ax: WCSAxes) -> None:
        """Add the scale bar to a finder chart.

        ax: `~astropy.visualization.wcsaxes.WCSAxes`
            WCS axes object.
        """
        # Add the horizontal line
        pixel_scale = pixel_scales(self._wcs)[0]
        parameters = self._parameters(self._minimum_length, pixel_scale)
        x_min = ax.get_xlim()[0] + self._left_edge[0]
        x_max = x_min + parameters.pixels
        y_mid = ax.get_ylim()[0] + self._left_edge[1]
        horizontal_path = Path([(x_min, y_mid), (x_max, y_mid)], closed=False)  # noqa
        horizontal_path_patch = PathPatch(horizontal_path, **self._kwargs)
        horizontal_path_patch.set_clip_on(False)
        ax.add_patch(horizontal_path_patch)

        # Add the vertical lines
        y_min = y_mid - 10
        y_max = y_mid + 10
        left_vertical_path = Path(
            [(x_min, y_min), (x_min, y_max)], closed=False  # noqa
        )  # noqa
        left_vertical_path_patch = PathPatch(left_vertical_path, **self._kwargs)
        left_vertical_path_patch.set_clip_on(False)
        ax.add_patch(left_vertical_path_patch)
        right_vertical_path = Path(
            [(x_max, y_min), (x_max, y_max)], closed=False  # noqa
        )  # noqa
        right_vertical_path_patch = PathPatch(right_vertical_path, **self._kwargs)
        right_vertical_path_patch.set_clip_on(False)
        ax.add_patch(right_vertical_path_patch)

        # Add the text
        text_kwargs = self._kwargs.copy()
        del text_kwargs["edgecolor"]
        del text_kwargs["facecolor"]
        del text_kwargs["linewidth"]
        text_kwargs["color"] = self._kwargs["facecolor"]
        x_mid = (x_min + x_max) / 2
        text = f"{parameters.angle:.1f} {parameters.units}"
        ax.text(
            x_mid,
            y_mid - 20,
            text,
            fontsize=10,
            horizontalalignment="center",
            **text_kwargs,
        )

    def rotate(self, pivot: SkyCoord, angle: Angle) -> "Annotation":
        """Rotate this annotation around a pivot and return the result.

        The rotation angle is an angle on the sky, measured from north to east.

        Parameters
        ----------
        pivot: `~astropy.coordinates.SkyCoord`
            Point around which to rotate the annotation.
        angle: `~astropy.coordinates.Angle`
            Angle of rotation, measured from north towards the east.

        Returns
        -------
        `~imephu.annotation.Annotation`
            The annotation resulting from the rotation.
        """
        raise NotImplementedError()

    def translate(self, displacement: Angle) -> "Annotation":
        """
        Move this annotation along a displacement vector and return the result.

        Not implemented.

        Parameters
        ----------
        displacement: 2D array of angles
            The displacement by which to move the annotation.

        Returns
        -------
        `~imephu.annotation.Annotation`
            The annotation resulting from the translation.
        """
        raise NotImplementedError()

    @staticmethod
    def _parameters(minimum_length: float, pixel_scale: Angle) -> _ScaleBarParameters:
        minimum_angle = minimum_length * pixel_scale
        if minimum_angle <= 50 * u.arcsec:
            minimum_angle_arcsec = minimum_angle.to_value(u.arcsec)
            angle_value = ScaleBarLineAnnotation._preferred_length_value(
                minimum_angle_arcsec
            )
            angle = angle_value * u.arcsec
            units = "arcsec"
        else:
            minimum_angle_arcmin = minimum_angle.to_value(u.arcmin)
            angle_value = ScaleBarLineAnnotation._preferred_length_value(
                minimum_angle_arcmin
            )
            angle = angle_value * u.arcmin
            units = "arcmin"
        pixels = float(angle / pixel_scale)
        return _ScaleBarParameters(pixels=pixels, angle=angle_value, units=units)

    @staticmethod
    def _preferred_length_value(minimum_length: float) -> float:
        """
        Return the preferred length given a minimum length value.

        The preferred value p is chosen as follows for a minimum length l_min.

        1. If the l_min is less than or equal to 1, p is equal to 1.
        2. Otherwise, if l_min is less than or equal to 1, p is l_min rounded up to the
           nearest integer.
        3. Otherwise, p is l_min rounded up to the next integer divisible by 10.

        Parameters
        ----------
        minimum_length: float
            Minimum scale bar length.

        Returns
        -------
        float
            The preferred scale bar length.

        """
        if minimum_length <= 1:
            return 1
        if minimum_length <= 10:
            return math.ceil(minimum_length)

        normalizing_factor = math.pow(10, math.floor(math.log10(minimum_length)))
        minimum_length /= normalizing_factor
        minimum_length = math.ceil(minimum_length)
        minimum_length *= normalizing_factor
        return minimum_length



from typing import Optional

from astropy import units as u
from astropy.coordinates import Angle, SkyCoord
from astropy.wcs import WCS
from imephu.annotation.general import (
    CrosshairsAnnotation,
    EmptyAnnotation,
    GroupAnnotation,
    RectangleAnnotation,
    TextAnnotation,
)
from imephu.annotation.general.scale_bar import ScaleBarLineAnnotation
from imephu.salt.annotation import rss, salticam


def title_annotation(
    target: str, proposal_code: str, pi_family_name: str, wcs: WCS
) -> TextAnnotation:
    """Return a text annotation with the title to the finder chart.

    The title is of the form "target name (proposal code; surname of the Principal
    Investigator)", such as "Magrathea (2022-1-SCI-042; Adams)".

    Parameters
    ----------
    target: str
        The name of the target for which the finder chart is intended.
    proposal_code: str
        The proposal code for which the finder chart is intended.
    pi_family_name: str
        The family name of the Principal Investigator.
    wcs: `~astropy.wcs.WCS`
        WCS object.

    Returns
    -------
    `~imephu.annotation.general.TextAnnotation`
        The title annotation.
    """
    title_text = f"{target} ({proposal_code}; {pi_family_name})"
    return TextAnnotation(
        (0.5, 1.03),
        title_text,
        wcs=wcs,
        color="black",
        horizontalalignment="center",
        style="italic",
        weight="bold",
        size="large",
        clip_on=False,
    )


def directions_annotation(fits_center: SkyCoord, wcs: WCS) -> GroupAnnotation:
    """Return an annotation for the east and north direction.

    Parameters
    ----------
    fits_center: `~astropy.coordinates.SkyCoord`
        The central position of the finder chart, in right ascension and declination.
    wcs: `~astropy.wcs.WCS`
        WCS object.

    Returns
    -------
    `~astropy.annotation.general.GroupAnnotation`
        The annotation for the east and north direction.
    """
    east_annotation = TextAnnotation(
        fits_center,
        "E",
        wcs=wcs,
        color=(0, 0.5, 1),
        horizontalalignment="right",
        style="italic",
        weight="bold",
        size="large",
    ).translate(Angle((4.8, 0) * u.arcmin))
    north_annotation = TextAnnotation(
        fits_center,
        "N",
        wcs=wcs,
        color=(0, 0.5, 1),
        style="italic",
        weight="bold",
        size="large",
    ).translate(Angle((0, 4.8) * u.arcmin))
    return GroupAnnotation([east_annotation, north_annotation])


def position_angle_annotation(
    position_angle: Angle, automated_position_angle: bool, wcs: WCS
) -> TextAnnotation:
    """Return a text annotation with the position angle.

    The text of the annotation is "PA: angle (auto)", where angle is a value in degrees,
    given with one fractional digit, and where "(auto)" is only included if the
    ``automated`` parameter is ``True``.

    Parameters
    ----------
    position_angle: `~astropy.coordinates.Angle`
        The position angle, as an angle on the sky from north to east.
    automated_position_angle: bool
        Whether the position angle has been calculated automatically.
    wcs: `~astropy.wcs.WCS`
        WCS object.

    Returns
    -------
    `~imephu.annotation.general.TextAnnotation`
        The annotation for the position angle.
    """
    text = f"PA = {position_angle.to_value(u.degree):.1f}"
    if automated_position_angle:
        text += " (auto)"
    return TextAnnotation(
        (1, -0.068),
        text,
        wcs=wcs,
        color="black",
        horizontalalignment="right",
        verticalalignment="baseline",
        style="italic",
        weight="bold",
        size="large",
        clip_on=False,
    )


def survey_annotation(survey: str, wcs: WCS) -> TextAnnotation:
    """Return a text annotation with the survey name.

    Parameters
    ----------
    survey: str
        The survey name.
    wcs: `~astropy.wcs.WCS`
        WCS object.

    Returns
    -------
    `~imephu.annotation.general.TextAnnotation`
        The annotation with the survey name.
    """
    return TextAnnotation(
        (0, -0.068),
        survey,
        wcs=wcs,
        color="black",
        horizontalalignment="left",
        verticalalignment="baseline",
        style="italic",
        weight="bold",
        size="large",
        clip_on=False,
    )


def magnitude_range_annotation(
    bandpass: str,
    min_magnitude: float,
    max_magnitude: float,
    wcs: WCS,
) -> TextAnnotation:
    """Return a text annotation with the magnitude range.

    The magnitude range is given as a string of the form "bandpass = min magnitude -
    max magnitude", where the magnitudes have one fractional digit. If the difference
    between the minimum and maximum magnitude is less than 0.1, only the maximum
    magnitude is included.

    Parameters
    ----------
    bandpass: `str`
        Bandpass for which the magnitude range is given.
    min_magnitude: `float`
        Minimum (brightest) magnitude.
    max_magnitude: `float`
        Maximum (faintest) magnitude.
    fits_center: `~astropy.coordinates.SkyCoord`
        Center position of the finder chart, as a right ascension and declination.
    wcs: `~astropy.wcs.WCS`
        WCS object.

    Returns
    -------
    `~imephu.annotation.general.TextAnnotation`
        The magnitude range annotation.
    """
    if abs(max_magnitude - min_magnitude) > 0.09999:
        text = f"{bandpass} = {min_magnitude:.1f} - {max_magnitude:.1f}"
    else:
        text = f"{bandpass} = {max_magnitude:.1f}"
    return TextAnnotation(
        (0.5, 0.02),
        text,
        wcs,
        color=(0, 0.5, 1),
        style="italic",
        weight="bold",
        size="large",
        horizontalalignment="center",
        verticalalignment="baseline",
    )


def slot_annotation(
    center: SkyCoord, position_angle: Angle, wcs: WCS
) -> RectangleAnnotation:
    """Return an annotation showing the slot.

    Parameters
    ----------
    center: `~astropy.coordinates.SkyCoord`
        The center of the slot on the sky, in right ascension and declination.
    position_angle: `~astropy.coordinates.Angle`
        The position angle as an angle on the sky, measured from north to east.
    wcs: `~astropy.wcs.WCS`
        WCS object.

    Returns
    -------
    `~imephu.annotation.general.RectangleAnnotation`
        The slot annotation.
    """
    return RectangleAnnotation(
        center=center,
        width=1 * u.arcmin / 3,
        height=10 * u.arcmin,
        wcs=wcs,
        edgecolor="red",
        alpha=0.5,
        linewidth=2,
    ).rotate(center, position_angle + 90 * u.deg)


def base_annotations(
    target: str,
    proposal_code: str,
    pi_family_name: str,
    position_angle: Angle,
    automated_position_angle: bool,
    survey: Optional[str],
    fits_center: SkyCoord,
    wcs: WCS,
) -> GroupAnnotation:
    """Return a group containing all the base annotations for SALT finder charts.

    The base annotations include the title, directions, position angle, survey name and
    the RSS and Salticam fields of view.

    Parameters
    ----------
    target: str
        The name of the target for which the finder chart is intended.
    proposal_code: str
        The proposal code for which the finder chart is intended.
    pi_family_name: str
        The family name of the Principal Investigator.
    position_angle: `~astropy.coordinates.Angle`
        The position angle, as an angle on the sky from north to east.
    automated_position_angle: bool
        Whether the position angle has been calculated automatically.
    survey: str, optional
        The image survey name.
    fits_center: `~astropy.coordinates.SkyCoord`
        The central position of the finder chart, in right ascension and declination
    wcs: `~astropy.wcs.WCS`
        WCS object.

    Returns
    -------
    `~imephu.annotation.general.GroupAnnotation`
        An annotation with all the SALT base annotations.
    """
    return GroupAnnotation(
        [
            title_annotation(
                target=target,
                proposal_code=proposal_code,
                pi_family_name=pi_family_name,
                wcs=wcs,
            ),
            survey_annotation(survey=survey, wcs=wcs) if survey else EmptyAnnotation(),
            position_angle_annotation(
                position_angle=position_angle,
                automated_position_angle=automated_position_angle,
                wcs=wcs,
            ),
            directions_annotation(fits_center=fits_center, wcs=wcs),
            _crosshairs_annotation(fits_center=fits_center, wcs=wcs),
            salticam.field_of_view_annotation(fits_center=fits_center, wcs=wcs),
            rss.field_of_view_annotation(fits_center=fits_center, wcs=wcs),
            ScaleBarLineAnnotation(left_edge=(50, 50), minimum_length=100, wcs=wcs),
        ]
    )


def _crosshairs_annotation(fits_center: SkyCoord, wcs: WCS) -> CrosshairsAnnotation:
    return CrosshairsAnnotation(
        center=fits_center, size=8 * u.arcmin, wcs=wcs, color="green"
    )



import urllib.parse
from io import BytesIO
from typing import BinaryIO, Callable, Dict, NamedTuple, Protocol

import requests
from astropy import units as u
from astropy.coordinates import Angle, SkyCoord


class SurveyError(BaseException):
    """An exception arising when querying a sky survey."""

    pass


class _SurveyDetails(NamedTuple):
    identifier: str
    is_covering_position: Callable[[SkyCoord], bool]


def _always_true(x: SkyCoord) -> bool:
    return True


_DSS_DETAILS: Dict[str, _SurveyDetails] = {
    "POSS2/UKSTU Red": _SurveyDetails("poss2ukstu_red", _always_true),
    "POSS2/UKSTU Blue": _SurveyDetails("poss2ukstu_blue", _always_true),
    "POSS2/UKSTU IR": _SurveyDetails("poss2ukstu_ir", _always_true),
    "POSS1 Red": _SurveyDetails("poss1_red", lambda p: p.dec.to_value(u.deg) > -30),
    "POSS1 Blue": _SurveyDetails("poss1_blue", lambda p: p.dec.to_value(u.deg) > -30),
    "Quick-V": _SurveyDetails("quickv", lambda p: p.dec.to_value(u.deg) > 6),
    "HST Phase2 (GSC2)": _SurveyDetails("phase2_gsc2", _always_true),
    "HST Phase2 (GSC1)": _SurveyDetails(
        "phase2_gsc1",
        lambda p: p.dec.to_value(u.deg) < -20 or p.dec.to_value(u.deg) > 6,
    ),
}


_SKYVIEW_DETAILS = {
    "2MASS-H": _SurveyDetails("2mass-h", _always_true),
    "2MASS-J": _SurveyDetails("2mass-j", _always_true),
    "2MASS-K": _SurveyDetails("2mass-k", _always_true),
}


class SkySurvey(Protocol):
    """A sky survey from which fits files can be requested."""

    def url(self, survey: str, fits_center: SkyCoord, size: Angle) -> str:
        """Return the URL for loading the FITS file for a given position.

        See `load_fits` for the available surveys.

        Parameters
        ----------
        survey: `str`
            The name of the survey to query for the FITS file.
        fits_center: `~astropy.coordinates.SkyCoord`
            The center position of the loaded FITS file.
        size: `~astropy.coordinates.Angle`
            The width and height of the FITS image, as an angle on the sky.

        Returns
        -------
        str
            The URL.
        """
        raise NotImplementedError

    def load_fits(self, survey: str, fits_center: SkyCoord, size: Angle) -> BinaryIO:
        """Load a FITS file for a given position from the sky survey.

        The size of the returned FITS file is determined by the class implementing this
        protocol (or the web service used).

        Parameters
        ----------
        survey: `str`
            The name of the survey to query for the FITS file.
        fits_center: `~astropy.coordinates.SkyCoord`
            The center position of the loaded FITS file.
        size: `~astropy.coordinates.Angle`
            The width and height of the FITS image, as an angle on the sky.

        Returns
        -------
        binary stream
            The FITS file.
        """
        raise NotImplementedError

    def is_covering_position(self, survey: str, position: SkyCoord) -> bool:
        """Return whether a position is covered by the sky survey.

        Parameters
        ----------
        survey: `str`
            The name of the survey.
        position: `~astropy.coordinates.SkyCoord`
            The position.

        Returns
        -------
        bool
            True if the position is covered  by the survey, False otherwise.
        """
        raise NotImplementedError


class DigitizedSkySurvey(SkySurvey):
    """A class for loading FITS files from the Digitized Sky Survey (DSS).

    See the `DSS archive website <https://archive.stsci.edu/dss/index.html>` for more
    details.
    """

    def __init__(self) -> None:
        self._survey_details: Dict[str, _SurveyDetails] = {
            key.lower(): value for key, value in _DSS_DETAILS.items()
        }

    def url(self, survey: str, fits_center: SkyCoord, size: Angle) -> str:
        """Return the URL for loading the FITS file for a given position.

        See `load_fits` for the available surveys.

        Parameters
        ----------
        survey: `str`
            The name of the survey to query for the FITS file.
        fits_center: `~astropy.coordinates.SkyCoord`
            The center position of the loaded FITS file.
        size: `~astropy.coordinates.Angle`
            The width and height of the FITS image, as an angle on the sky.

        Returns
        -------
        str
            The URL.
        """
        url = "https://archive.stsci.edu/cgi-bin/dss_search"
        params = {
            "v": self._survey_identifier(survey),
            "r": fits_center.ra.to_value(u.deg),
            "d": fits_center.dec.to_value(u.deg),
            "e": "J2000",
            "h": size.to_value(u.arcmin),
            "w": size.to_value(u.arcmin),
            "f": "fits",
            "c": "none",
        }
        return url + "?" + urllib.parse.urlencode(params)

    def load_fits(self, survey: str, fits_center: SkyCoord, size: Angle) -> BinaryIO:
        """Load a FITS file for a given position from the sky survey.

        The following surveys can be queried for FITS images:

        - POSS2/UKSTU Red
        - POSS2/UKSTU Blue
        - POSS2/UKSTU IR
        - POSS1 Red
        - POSS1 Blue
        - Quick-V
        - HST Phase2 (GSC2)
        - HST Phase2 (GSC1)

        See the `DSS survey page <http://gsss.stsci.edu/SkySurveys/Surveys.htm>` for
        details about these surveys.

        Parameters
        ----------
        survey: `str`
            The name of the survey to query for the FITS file.
        fits_center: `~astropy.coordinates.SkyCoord`
            The center position of the loaded FITS file.
        size: `~astropy.coordinates.Angle`
            The width and height of the FITS image, as an angle on the sky.

        Returns
        -------
        binary stream
            The FITS file.
        """
        try:
            response = requests.get(self.url(survey, fits_center, size))
            if response.status_code != 200:
                raise SurveyError("No FITS file could be loaded.")
            return BytesIO(response.content)
        except BaseException as e:
            raise SurveyError("No FITS file could be loaded.") from e

    def is_covering_position(self, survey: str, position: SkyCoord) -> bool:
        """Return whether a position is covered by the sky survey.

        Parameters
        ----------
        survey: `str`
            The name of the survey.
        position: `~astropy.coordinates.SkyCoord`
            The position.

        Returns
        -------
        bool
            True if the position is covered  by the survey, False otherwise.
        """
        return self._survey_details[survey.lower()].is_covering_position(position)

    def _survey_identifier(self, survey: str) -> str:
        if survey.lower() not in self._survey_details:
            raise ValueError(f"Unknown survey: {survey}")

        return self._survey_details[survey.lower()].identifier


class SkyView(SkySurvey):
    """A class for loading FITS files from SkyView.

    See the `SkyView site <https://skyview.gsfc.nasa.gov/current/cgi/titlepage.pl>` for
    more details about SkyView.

    Parameters
    ----------
    pixels: `int`, default: 300
        The image size in pixels.
    """

    def __init__(self, pixels: int = 300) -> None:
        self._pixels = pixels
        self._survey_details: Dict[str, _SurveyDetails] = {
            key.lower(): value for key, value in _SKYVIEW_DETAILS.items()
        }

    def url(self, survey: str, fits_center: SkyCoord, size: Angle) -> str:
        """Return the URL for loading the FITS file for a given position.

        See `load_fits` for the available surveys.

        Parameters
        ----------
        survey: `str`
            The name of the survey to query for the FITS file.
        fits_center: `~astropy.coordinates.SkyCoord`
            The center position of the loaded FITS file.
        size: `~astropy.coordinates.Angle`
            The width and height of the FITS image, as an angle on the sky.

        Returns
        -------
        str
            The URL.
        """
        url = "https://skyview.gsfc.nasa.gov/current/cgi/runquery.pl"
        params = {
            "Position": f"{fits_center.ra.to_value(u.deg)}, "
            f"{fits_center.dec.to_value(u.deg)}",
            "Survey": self._survey_identifier(survey),
            "Coordinates": "ICRS",
            "Return": "FITS",
            "Pixels": self._pixels,
            "Size": size.to_value(u.deg),
        }
        return url + "?" + urllib.parse.urlencode(params)

    def load_fits(self, survey: str, fits_center: SkyCoord, size: Angle) -> BinaryIO:
        """Load a FITS file for a given position from the sky survey.

        The following subset of surveys supported by SkyView can be queried:

        - 2MASS-H
        - 2MASS-J
        - 2MASS-K

        See the `SkyView survey page
        <https://skyview.gsfc.nasa.gov/current/cgi/survey.pl>` for details about these
        surveys.

        Parameters
        ----------
        survey: `str`
            The name of the survey to query for the FITS file.
        fits_center: `~astropy.coordinates.SkyCoord`
            The center position of the loaded FITS file.
        size: `~astropy.coordinates.Angle`
            The width and height of the FITS image, as an angle on the sky.

        Returns
        -------
        binary stream
            The FITS file.
        """
        try:
            response = requests.get(self.url(survey, fits_center, size))
            if response.status_code != 200:
                raise SurveyError("No FITS file could be loaded.")
            return BytesIO(response.content)
        except BaseException as e:
            raise SurveyError("No FITS file could be loaded.") from e

    def is_covering_position(self, survey: str, position: SkyCoord) -> bool:
        """Return whether a position is covered by the sky survey.

        Parameters
        ----------
        survey: `str`
            The name of the survey.
        position: `~astropy.coordinates.SkyCoord`
            The position.

        Returns
        -------
        bool
            True if the position is covered  by the survey, False otherwise.
        """
        return self._survey_details[survey.lower()].is_covering_position(position)

    def _survey_identifier(self, survey: str) -> str:
        if survey.lower() not in self._survey_details:
            raise SurveyError(f"Unknown survey: {survey}")

        return self._survey_details[survey.lower()].identifier


def url(survey: str, fits_center: SkyCoord, size: Angle) -> str:
    """Return the URL for requesting a FITS image from a sky survey.

    In principle, this function is equivalent to the `url` methods of the
    `SkySurvey` implementations. For example,

    .. code:: python

       from astropy import units as u
       from astropy.coordinates import SkyCoord

       url("POSS2/UKSTU Red",
           SkyCoord(ra=120 * u.deg, dec=-30 * u.deg),
           10 * u.arcmin)

    is the same as

    .. code:: python

       survey = DigitizedSkySurvey()
       survey.url("POSS2/UKSTU Red",
                  SkyCoord(ra=120 * u.deg, dec=-30 * u.deg),
                  10 * u.arcmin)

    The advantage of this function is that you don't have to remember which survey
    belongs to which `SkySurvey` implementation.

    See the `SkySurvey` implementations for the supported surveys.

    Parameters
    ----------
    survey: `str`
        The name of the survey to query for the FITS file.
    fits_center: `~astropy.coordinates.SkyCoord`
        The center position of the loaded FITS file.
    size: `~astropy.coordinates.Angle`
        The width and height of the FITS image, as an angle on the sky.

    Returns
    -------
    str
        The URL.
    """
    if survey.lower() in [s.lower() for s in _DSS_DETAILS.keys()]:
        survey_: SkySurvey = DigitizedSkySurvey()
    elif survey.lower() in [s.lower() for s in _SKYVIEW_DETAILS.keys()]:
        survey_ = SkyView(pixels=700)
    else:
        raise ValueError(f"Unknown survey: {survey}")

    return survey_.url(survey, fits_center, size)


def load_fits(survey: str, fits_center: SkyCoord, size: Angle) -> BinaryIO:
    """Request a FITS image from a sky survey.

    In principle, this function is equivalent to the `load_fits` methods of the
    `SkySurvey` implementations. For example,

    .. code:: python

       from astropy import units as u
       from astropy.coordinates import SkyCoord

       load_fits("POSS2/UKSTU Red",
                 SkyCoord(ra=120 * u.deg, dec=-30 * u.deg),
                 10 * u.arcmin)

    is the same as

    .. code:: python

       survey = DigitizedSkySurvey()
       survey.load_fits("POSS2/UKSTU Red",
                        SkyCoord(ra=120 * u.deg, dec=-30 * u.deg),
                        10 * u.arcmin)

    The advantage of this function is that you don't have to remember which survey
    belongs to which `SkySurvey` implementation.

    See the `SkySurvey` implementations for the supported surveys.

    Parameters
    ----------
    survey: `str`
        The name of the survey to query for the FITS file.
    fits_center: `~astropy.coordinates.SkyCoord`
        The center position of the loaded FITS file.
    size: `~astropy.coordinates.Angle`
        The width and height of the FITS image, as an angle on the sky.

    Returns
    -------
    binary stream
        The FITS file.
    """
    response = requests.get(url(survey, fits_center, size))
    if response.status_code != 200:
        raise SurveyError("No FITS file could be loaded.")
    return BytesIO(response.content)


def is_covering_position(survey: str, position: SkyCoord) -> bool:
    """Return whether a position is covered by a sky survey.

    In principle, this function is equivalent to the `is_covering_position` methods of
    the `SkySurvey` implementations. For example,

    .. code:: python

       from astropy import units as u
       from astropy.coordinates import SkyCoord

       is_covering_position("POSS2/UKSTU Red",
                            SkyCoord(ra=120 * u.deg, dec=-30 * u.deg))

    is the same as

    .. code:: python

       survey = DigitizedSkySurvey()
       survey.is_covering_position("POSS2/UKSTU Red",
                                   SkyCoord(ra=120 * u.deg, dec=-30 * u.deg))

    The advantage of this function is that you don't have to remember which survey
    belongs to which `SkySurvey` implementation.

    See the `SkySurvey` implementations for the supported surveys.

    Parameters
    ----------
    survey: `str`
        The name of the survey.
    position: `~astropy.coordinates.SkyCoord`
        The position.

    Returns
    -------
    binary stream
        The FITS file.
    """
    if survey.lower() in [s.lower() for s in _DSS_DETAILS.keys()]:
        survey_: SkySurvey = DigitizedSkySurvey()
    elif survey.lower() in [s.lower() for s in _SKYVIEW_DETAILS.keys()]:
        survey_ = SkyView(pixels=700)
    else:
        raise ValueError(f"Unknown survey: {survey}")
    return survey_.is_covering_position(survey, position)



from __future__ import annotations

import bisect
import json
import os
import warnings
from copy import deepcopy
from datetime import datetime
from io import BytesIO
from typing import (
    Any,
    BinaryIO,
    Callable,
    Dict,
    Generator,
    List,
    Optional,
    Tuple,
    Union,
)

import matplotlib.pyplot as plt
from astropy import units as u
from astropy.coordinates import Angle, SkyCoord
from astropy.io import fits
from astropy.visualization.interval import AsymmetricPercentileInterval
from astropy.visualization.mpl_normalize import simple_norm
from astropy.visualization.wcsaxes.core import WCSAxesSubplot
from astropy.wcs import WCS, FITSFixedWarning
from matplotlib.figure import Figure
from PyPDF2 import PdfReader, PdfWriter

import imephu
from imephu.annotation import Annotation
from imephu.geometry import pixel_scales
from imephu.service.survey import load_fits
from imephu.utils import Ephemeris

"""A finder chart."""


class FinderChart:
    """A finder chart for an astronomical observation.

    The finder chart is generated from a FITS file. By default the finder chart just
    shows the region with an overlaid coordinate grid. The axes show WCS coordinates.
    Use the `add_annotation` method to add more content to the finder chart.

    By default, the size of the finder chart is the same as that of the FITS file. You
    may set the `max_size` property to choose a non-default maximum size.

    You can display the finder chart on the screen or save it as a file.

    This class uses Matplotlib for generating the finder chart.

    .. warning:: Be careful when working with several finder charts as all the finder
       charts use the same Matplotlib figure.

    Parameters
    ----------
    name: `str`, `path-like` or `binary file-like`
        FITS file to display.
    max_size: `~astropy.coordinates.Angle`
        Maximum size of the generated finder chart.
    """

    def __init__(
        self,
        name: Union[str, BinaryIO, os.PathLike[Any]],
        max_size: Optional[Angle] = None,
    ):
        self._hdu = fits.open(name)[0]
        # The FITS data is read in only when it is needed. To avoid trying to read from
        # a closed stream later on, we thus force the data to be read in immediately.
        self._data = self._hdu.data
        self._metadata: Dict[str, Any] = dict()
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=FITSFixedWarning)
            self._wcs = WCS(self._hdu)
        self._annotations: List[Annotation] = []
        self._max_size = max_size

    @staticmethod
    def from_survey(survey: str, fits_center: SkyCoord, size: Angle) -> "FinderChart":
        """Create a finder chart from a sky survey FITS image.

        Parameters
        ----------
        survey: `str`
            The name of the survey to query for the FITS file.
        fits_center: `~astropy.coordinates.SkyCoord`
            The center position of the loaded FITS file.
        size: `~astropy.coordinates.Angle`
            The width and height of the FITS file image, as an angle on the sky.
        """
        fits_image = load_fits(survey, fits_center, size)
        return FinderChart(fits_image)

    @staticmethod
    def for_time_interval(
        start: datetime,
        end: datetime,
        ephemerides: List[Ephemeris],
        max_track_length: Angle,
        create_finder_chart: Callable[[List[Ephemeris]], "FinderChart"],
    ) -> Generator[Tuple["FinderChart", Tuple[datetime, datetime]], None, None]:
        """Create finder charts for a time interval.

        This method is intended for non-sidereal targets, which may require multiple
        finder charts to cover their track over a time interval. It creates finder
        charts according to the following rules:

        * Taken together, the finder charts cover the whole time interval.
        * The track length on each finder chart does not exceed `max_track_length`.
        * Each finder chart is created using ``create_finder_chart``.

        The method is completely agnostic as to what a finder chart should look like;
        this decision is left completely to ``create_finder_chart``. In particular, if
        you need any annotations for the non-sidereal nature, it is up to
        ``create_finder_chart`` to provide them.

        The ``create_finder_chart`` function has to accept as its single argument the
        list of ephemerides to include on the finder chart.

        The finder charts are returned along with the time interval they cover. The time
        interval is also added as a tuple of `~datetime.datetime` values with the key
        ``valid_for`` to the finder chart's metadata.

        Parameters
        ----------
        start: `~datetime.datetime`
            The start time of the interval to be covered by the finder charts. This must
            be a timezone-aware datetime.
        end: `~datetime.datetime`
            The end time of the interval to be covered by the finder charts. This must
            be a timezone-aware datetime.
        ephemerides: list of `~imephu.utils.Ephemeris`
            The list of ephemerides. The time interval from ``start`` to ``end`` must be
            fully covered by the ephemerides.
        max_track_length: float
            The maximum length a track may have on a finder chart, as an angle on the
            sky.
        create_finder_chart: function
            The function for creating a finder chart from a list of ephemerides.

        Yields
        ------
        tuple of a `~imephu.finder_chart.FinderChart` and a datetime interval
            The generated finder charts along with the time intervals they cover.
        """
        # The start and end time must be timezone-aware
        if start.tzinfo is None or start.tzinfo.utcoffset(None) is None:
            raise ValueError("The start time must be timezone-aware.")
        if end.tzinfo is None or end.tzinfo.utcoffset(None) is None:
            raise ValueError("The end time must be timezone-aware.")

        if start >= end:
            raise ValueError("The start time must be earlier than the end time.")

        if start < ephemerides[0].epoch:
            raise ValueError("The start time must not be earlier than the first epoch.")
        if end > ephemerides[-1].epoch:
            raise ValueError("The end time must not be later than the last epoch.")

        if max_track_length.to_value(u.arcmin) <= 0:
            raise ValueError("The maximum track length must be positive.")

        # Find the smallest interval covering the time interval
        all_times = [e.epoch for e in ephemerides]
        start_index = bisect.bisect_right(all_times, start) - 1
        end_index = bisect.bisect(all_times, end)
        if end_index > 0 and all_times[end_index - 1] == end:
            end_index -= 1

        # This should never happen, but let's rule out intervals with zero length
        if start_index == end_index:
            raise ValueError("The interval must have a positive length")

        # Split the ephemerides so that the maximum track length isn't exceeded for each
        # group. We assume a linear path, so that the track length is equal to the angle
        # between the first and last position.
        current_group = [ephemerides[start_index]]
        groups = [current_group]
        for i in range(start_index + 1, end_index + 1):
            next_ephemeris = ephemerides[i]
            # Subsequent positions must not be more than max_track_length apart
            if (
                current_group[-1].position.separation(next_ephemeris.position)
                > max_track_length
            ):
                raise ValueError(
                    "The maximum track length on a finder chart is exceeded."
                )
            track_length = current_group[0].position.separation(next_ephemeris.position)
            if track_length <= max_track_length:
                current_group.append(next_ephemeris)
            else:
                current_group = [current_group[-1], next_ephemeris]
                groups.append(current_group)

        # Create the finder charts
        for group in groups:
            valid_for = (group[0].epoch, group[-1].epoch)
            finder_chart = create_finder_chart(group)
            finder_chart.add_metadata(
                "valid_for", [valid_for[0].isoformat(), valid_for[1].isoformat()]
            )
            yield finder_chart, valid_for

    @property
    def wcs(self) -> WCS:
        """Return a deep copy of the WCS object of the finder chart.

        Returns
        -------
        `~astropy.wcs.WCS`
            A deep copy of the WCS object.
        """
        return self._wcs.deepcopy()

    def add_annotation(self, annotation: Annotation) -> None:
        """Add an annotation to the finder chart.

        Annotations will be plotted onto the finder chart in the order they have been
        added. So, for example, the annotation added last will be output on top of all
        the other annotations.
        """
        self._annotations.append(annotation)

    @property
    def metadata(self) -> Dict[str, Any]:
        """
        Return the metadata for the finder chart.

        Metadata can  be added with the `add_metadata` method. A shallow copy of the
        metadata is returned.

        Returns
        -------
        dict
            The metadata of the finder chart.
        """
        return self._metadata.copy()

    def add_metadata(self, key: str, value: Any) -> None:
        """
        Add a key-value to the finder chart's metadata.

        If the key existrs in the metadata already, the existing value for the key is
        replaced.

        Parameters
        ----------
        key: `str`
            Key.
        value: `~typing.Any`
            Value.
        """
        self._metadata[key] = value

    @property
    def max_size(self) -> Optional[Angle]:
        return self._max_size

    @max_size.setter
    def max_size(self, new_max_size: Optional[Angle]) -> None:
        self._max_size = new_max_size

    def show(self) -> None:
        """Display the finder chart on the screen."""
        figure = self._create_plot()
        plt.show()
        plt.close(figure)

    def save(
        self,
        name: Union[str, BinaryIO, os.PathLike[str], os.PathLike[bytes]],
        format: Optional[str] = None,
    ) -> None:
        """Save the finder chart in a file.

        If `format` is not set, the file extension of `name` is used to figure out the
        file or, if there is no extension, Matplotlib's default is used. See
        Matplotlib's `matplotlib.pyplot.savefig` function for more details regarding
        the format.

        Parameters
        ----------
        name: `str`, `path-like` or `binary file-like`
            The file to which the finder chart is saved. An existing file is replaced.
        format: `str`, optional
            The format in which to store the finder chart.
        """
        figure = self._create_plot()
        if format and format.lower() == "pdf":
            pdf = BytesIO()
            plt.savefig(pdf, format=format, bbox_inches="tight")

            pdf = self._add_metadata(pdf)

            if hasattr(name, "write"):
                name.write(pdf.getvalue())
            else:
                with open(name, "wb") as f:
                    f.write(pdf.getvalue())
        else:
            plt.savefig(name, format=format, bbox_inches="tight")
        plt.close(figure)

    def _create_plot(self) -> Figure:
        figure = plt.figure(figsize=(10, 9))
        ax = plt.subplot(projection=self._wcs)

        self._add_fits_content(ax)

        # Enforce the maximum finder chart size. This needs to be done after the FITS
        # file has been added, as that sets the plot limits according to the image size.
        self._enforce_max_size(ax)

        self._update_axes(ax)

        ax.grid(True, color="blue", alpha=0.2)

        for annotation in self._annotations:
            annotation.add_to(ax)

        return figure

    def _add_fits_content(self, ax: WCSAxesSubplot) -> None:
        """
         Add content of the FITS file to the plot.

        The code has been adapted from APLpy.
        """
        cmap = plt.cm.get_cmap("gist_yarg")

        # value range in the image
        pmin = 0.25
        pmax = 99.75
        interval = AsymmetricPercentileInterval(pmin, pmax, n_samples=10000)
        vmin, vmax = interval.get_limits(self._hdu.data)
        vmin = -0.1 * (vmax - vmin) + vmin
        vmax = 0.1 * (vmax - vmin) + vmax

        normalizer = simple_norm(self._data, power=2, min_cut=vmin, max_cut=vmax)

        ax.imshow(
            self._hdu.data,
            cmap=cmap,
            interpolation="nearest",
            origin="lower",
            norm=normalizer,
            aspect="equal",
        )

    def _enforce_max_size(self, ax: WCSAxesSubplot) -> None:
        if self.max_size is None:
            return

        # Get the current plot limits
        x_min, x_max = ax.get_xlim()
        y_min, y_max = ax.get_ylim()

        # Get the angular size in x and y direction, and assume the finder chart size
        # to be the greater of the two
        x_scale, y_scale = pixel_scales(ax.wcs)
        x_size_pixels = x_max - x_min
        y_size_pixels = y_max - y_min
        x_size = x_size_pixels * x_scale
        y_size = y_size_pixels * y_scale
        size = max(x_size, y_size)

        # If we are within 5% of the maximum size, there is nothing to do
        if size < 1.05 * self.max_size:
            return

        # Update the limits in x direction
        shrink_factor = float(self.max_size / size)
        x_center = (x_min + x_max) / 2
        x_size_data_coords_new = shrink_factor * x_size_pixels
        x_min_new = x_center - x_size_data_coords_new / 2
        x_max_new = x_center + x_size_data_coords_new / 2
        ax.set_xlim(left=x_min_new, right=x_max_new)

        # Update the limits in y direction
        y_center = (y_min + y_max) / 2
        y_size_data_coords_new = shrink_factor * y_size_pixels
        y_min_new = y_center - y_size_data_coords_new / 2
        y_max_new = y_center + y_size_data_coords_new / 2
        ax.set_ylim(bottom=y_min_new, top=y_max_new)

    def _update_axes(self, ax: WCSAxesSubplot) -> None:
        axis_type_names = {
            "pos.eq.dec": "Dec (ICRS)",
            "pos.eq.ra": "RA (ICRS)",
        }
        axis_types = self._wcs.world_axis_physical_types
        x_axis_type = axis_types[0]
        y_axis_type = axis_types[1]
        if x_axis_type != "pos.eq.ra":
            raise ValueError(
                "Only pos.eq.ra is supported for the physical type of "
                "the first world axis"
            )
        if y_axis_type != "pos.eq.dec":
            raise ValueError(
                "Only pos.eq.dec is supported for the physical type of "
                "the second world axis"
            )

        x = ax.coords[0]
        y = ax.coords[1]
        x.set_axislabel(axis_type_names[x_axis_type])
        y.set_axislabel(axis_type_names[y_axis_type])

        x.display_minor_ticks(True)
        y.display_minor_ticks(True)

        x.set_ticks(size=7)
        y.set_ticks(size=7)

    def _add_metadata(self, pdf: BytesIO) -> BytesIO:
        """
        Add metadata to the pdf.

        The following metadata is added as a JSON file attachment named metadata.json to
        the given pdf.

        * All content of the finder chart's metadata dictionary.
        * imephu (and the version) as the property "author", unless the finder chart's
          metadata dictionary contains that property already.

        imephu (and the version) is also as the author metadata to the given pdf

        Parameters
        ----------
        pdf: `~io.BinaryIO`
            PDF.

        Returns
        -------
        `~io.BinaryIO`
            The PDF with the author metadata.
        """
        # Read the existing metadata
        pdf.seek(0)
        reader = PdfReader(pdf)
        meta = reader.metadata

        # Copy the pdf content
        writer = PdfWriter()
        for page in reader.pages:
            writer.add_page(page)

        # Add the metadata (and author)
        author = f"imephu ({imephu.__version__})"
        metadata = deepcopy(self.metadata)
        if "author" not in metadata:
            metadata["author"] = author
        writer.add_attachment("metadata.json", json.dumps(metadata).encode("utf-8"))

        # Collect the current metadata and add the author
        # Note that the slash in "/Author" is indeed required
        if meta:
            updated_meta = {str(k): str(v) for k, v in meta.items()}
        else:
            updated_meta = {}
        updated_meta["/Author"] = author
        writer.add_metadata(updated_meta)

        # Save the new pdf
        updated_pdf = BytesIO()
        writer.write(updated_pdf)
        updated_pdf.seek(0)

        return updated_pdf



from astropy import units as u
from astropy.coordinates import Angle, SkyCoord
from astropy.wcs import WCS
from imephu.annotation.general import (
    CircleAnnotation,
    GroupAnnotation,
    RectangleAnnotation,
    TextAnnotation,
)
from imephu.geometry import translate
from imephu.salt.utils import MosMask


def field_of_view_annotation(fits_center: SkyCoord, wcs: WCS) -> GroupAnnotation:
    """
    Return an annotation with the RSS field of view.

    Parameters
    ----------
    fits_center: `~astropy.coordinates.SkyCoord`
        The center of the finder chart, in right ascension and declination.
    wcs: `~astropy.wcs.WCS`
        WCS object.

    Returns
    -------
    `~imephu.annotation.general.GroupAnnotation`
        An annotation with the RSS field of view.
    """
    fov_annotation = CircleAnnotation(
        fits_center, 4 * u.arcmin, wcs=wcs, edgecolor="green"
    )
    label_position = translate(fits_center, (-2.9, 2.9) * u.arcmin)
    name_annotation = TextAnnotation(
        label_position,
        "RSS",
        wcs=wcs,
        style="italic",
        weight="bold",
        size="large",
        horizontalalignment="left",
        color=(0, 0, 1),
    )
    return GroupAnnotation([fov_annotation, name_annotation])


def longslit_annotation(
    fits_center: SkyCoord,
    slit_width: Angle,
    slit_height: Angle,
    position_angle: Angle,
    wcs: WCS,
) -> RectangleAnnotation:
    """Return an annotation showing a longslit.

    Parameters
    ----------
    fits_center: `~astropy.coordinates.SkyCoord`
        The central position of the finder chart, in right ascension and declination
    slit_width: `~astropy.coordinates.Angle`
        The width of the slit, as an angle on the sky.
    slit_height: `~astropy.coordinates.Angle`
        The height of the slit, as an angle on the sky.
    position_angle: `~astropy.coordinates.Angle`
        The position angle, as an angle on the sky measured from north to east.
    wcs: `~astropy.wcs.WCS`
        WCS object.

    Returns
    -------
    `imephu.annotation.general.RectangleAnnotation`
        The annotation showing the longslit.
    """
    return RectangleAnnotation(
        fits_center, slit_width, slit_height, wcs=wcs, edgecolor="red", alpha=0.5
    ).rotate(fits_center, position_angle)


def reference_star_annotation(reference_star: SkyCoord, wcs: WCS) -> CircleAnnotation:
    """Return a circle highlighting a reference star.

    Parameters
    ----------
    reference_star: `~astropy.coordinates.SkyCoord`
        The central position of the finder chart, in right ascension and declination
    wcs: `~astropy.wcs.WCS`
        WCS object.
    """
    return CircleAnnotation(
        reference_star,
        10 * u.arcsec,
        wcs,
        edgecolor="red",
        alpha=0.5,
    )


def mos_mask_annotation(
    mos_mask: MosMask, wcs: WCS, reference_star_box_width: Angle = 5 * u.arcsec
) -> GroupAnnotation:
    """Return the annotation for a MOS mask.

    The slits and boxes around the reference stars are included in the annotation.

    Parameters
    ----------
    mos_mask: `~imephu.salt.util.MosMask`
        The MOS mask.
    wcs: `~astropy.wcs.WCS`
        WCS object.
    reference_star_box_width: `~astropy.coordinates.Angle`, default: 5 arcseconds
        The width (and height) of the boxes around reference stars, as an angle on the
        sky.

    Returns
    -------
    `~imephu.annotation.general.GroupAnnotation`
        The annotation displaying the MOS mask.
    """
    mask_annotation = GroupAnnotation([])

    position_angle = mos_mask.position_angle
    for star in mos_mask.reference_stars:
        reference_star_annotation = RectangleAnnotation(
            center=star,
            width=reference_star_box_width,
            height=reference_star_box_width,
            wcs=wcs,
            edgecolor=(1, 1, 0),
            linewidth=2,
        ).rotate(star, position_angle)
        mask_annotation.add_item(reference_star_annotation)

    for slit in mos_mask.slits:
        slit_annotation = RectangleAnnotation(
            center=slit.center,
            width=slit.width,
            height=slit.height,
            wcs=wcs,
            edgecolor="red",
        ).rotate(slit.center, position_angle + slit.tilt)
        mask_annotation.add_item(slit_annotation)

    return mask_annotation



import numpy as np
from astropy import units as u
from astropy.coordinates import Angle, SkyCoord
from astropy.wcs import WCS
from imephu.annotation.general import (
    CircleAnnotation,
    CrosshairsAnnotation,
    GroupAnnotation,
    LinePathAnnotation,
)
from imephu.geometry import translate

_OUTLINE_COLOR = "red"


def bundles_annotation(
    finder_chart_center: SkyCoord,
    science_bundle_center: SkyCoord,
    bundle_separation: Angle,
    position_angle: Angle,
    wcs: WCS,
    include_fibers: bool = False,
) -> GroupAnnotation:
    """Return the annotation for NIR fiber bundles.

    The annotation contains the outline of the science fiber bundle (with its center
    marked by crosshairs) and the outlines of the sky fiber bundles.

    Parameters
    ----------
    finder_chart_center: `~astropy.coordinates.SkyCoord`
        The finder chart center, as a position on the sky, in right ascension and
        declination.
    science_bundle_center: `~astropy.coordinates.SkyCoord`
        The center position of the sky fiber bundle, as a position on the sky, in
        right ascension and declination.
    bundle_separation: ~astropy.coordinates.Angle`
        The separation between the science fiber bundle and the sky fiber bundles, as an
        angle on the sky. The separation is measured between the center of the science
        bundle and the midpoint of the line between the centers of the sky bundles.
    position_angle: `~astropy.coordinates.Angle`
        The position angle, as an angle on the sky measured from north to east.
    wcs:`astropy.wcs.WCS`
        WCS object.
    include_fibers: `bool`, optional, default False
        Whether to include the fiber outlines in the annotation.

    Returns
    -------
    `imephu.annotation.general.GroupAnnotation`
        The annotation for an NIR setup.
    """
    r = 0.65 * u.arcsec
    d_horizontal = 1.63 * u.arcsec
    d_vertical = 1.42 * u.arcsec
    bundle_separaration_arcsec = bundle_separation.to_value(u.arcsec)
    science_bundle = _science_bundle(
        r=r,
        d_horizontal=d_horizontal,
        d_vertical=d_vertical,
        center=finder_chart_center,
        wcs=wcs,
        include_fibers=include_fibers,
    ).rotate(finder_chart_center, 90 * u.deg + position_angle)
    sky_bundle = (
        _sky_bundle(
            r=r,
            d_horizontal=d_horizontal,
            d_vertical=d_vertical,
            center=finder_chart_center,
            wcs=wcs,
            include_fibers=include_fibers,
        )
        .rotate(finder_chart_center, 90 * u.deg)
        .translate((bundle_separaration_arcsec, 0) * u.arcsec)
        .rotate(finder_chart_center, position_angle)
    )
    annotation = GroupAnnotation([sky_bundle, science_bundle])

    target_offsets = finder_chart_center.spherical_offsets_to(science_bundle_center)

    return annotation.translate(target_offsets)


def _science_bundle(
    r: Angle,
    d_horizontal: Angle,
    d_vertical: Angle,
    center: SkyCoord,
    wcs: WCS,
    include_fibers: bool,
) -> GroupAnnotation:
    """
    Return an annotation showing the science bundle.

    The annotation includes the science bundle outline with crosshairs at the center
    and, optionally, the bundle fiber outlines.

    Parameters
    ----------
    r: `~astropy.coord.Angle`
        The fiber radius, as angle on the sky.
    d_horizontal: `~astropy.coord.Angle`
        The distance between adjacent fibers in the direction of right ascension, as an
        angle on the sky.
    d_vertical: `~astropy.coord.Angle`
        The distance between adjacent rows of fiber cdenters in the direction of
        declination, as an angle on the sky.
    center: `~astropy.coord.SkyCoords`
        The position of the bundle center, as right ascension and declination.
    wcs: `~astropy.wcs.WCS`
        WCS object.
    include_fibers: `bool`
        Whether to include the fiber outlines in the annotation.

    Returns
    -------
    `~imephu.annotation.general.GroupAnnotation`
        The annotation.
    """
    annotation = GroupAnnotation([])
    if include_fibers:
        annotation.add_item(
            _science_fiber_outlines(
                r=r,
                d_horizontal=d_horizontal,
                d_vertical=d_vertical,
                center=center,
                wcs=wcs,
            )
        )
    annotation.add_item(
        _science_bundle_outline(
            r=r,
            d_horizontal=d_horizontal,
            d_vertical=d_vertical,
            center=center,
            wcs=wcs,
        )
    )
    annotation.add_item(
        CrosshairsAnnotation(
            center=center, size=10 * u.arcsec, wcs=wcs, color=_OUTLINE_COLOR
        )
    )
    return annotation


def _science_bundle_outline(
    r: Angle,
    d_horizontal: Angle,
    d_vertical: Angle,
    center: SkyCoord,
    wcs: WCS,
    include_fibers: bool = False,
) -> LinePathAnnotation:
    # For simplicity, we first assume that the bundle center is located at (0, 0).
    # We start by establishing one of the slanting lines. We make use id the fact that
    # the line is parallel to the line through the centers of the fibers to which it is
    # tangential.
    dh = d_horizontal.to_value(u.arcsec)
    dv = d_vertical.to_value(u.arcsec)
    fiber_center = np.array([-8.5 * dh, 2 * dv]) * u.arcsec

    # m is the slope of the line. The vector n from the fiber center to the line is
    # perpendicular to the line.
    m = dv / (dh / 2)
    dr = np.array([-dv, dh / 2]) * u.arcsec
    n = dr / np.linalg.norm(dr)

    # The distance between the fiber center and the line is the fiber radius, and with
    # this in mind we can calculate the y offset c of the line.
    p = fiber_center + r * n
    c = p[1] - m * p[0]

    # The first end point of the line is the intersection of the adjacent horizontal
    # line and y = m x + c. In the following, gap_vertical is the vertical distance
    # between circles in neighbouring rows.
    gap_vertical = d_vertical - 2 * r
    y1 = 6.5 * d_vertical - gap_vertical / 2
    x1 = (y1 - c) / m
    vertices = []
    vertices.append((x1, y1))

    # The second end point of the line is the intersection of the adjacent vertical line
    # and y = m x + c. In the following, gap_horizontal is the horizontal distance
    # between adjacent circles.
    gap_horizontal = d_horizontal - 2 * r
    x2 = -9 * d_horizontal + gap_horizontal / 2
    y2 = m * x2 + c
    vertices.append((x2, y2))

    # The next four vertices are just end points of horizontal and vertical lines.
    x3 = x2
    y3 = 2 * d_vertical - r
    vertices.append((x3, y3))

    x4 = -8 * d_horizontal - r
    y4 = y3
    vertices.append((x4, y4))

    x5 = x4
    y5 = r
    vertices.append((x5, y5))

    x6 = x2
    y6 = y5
    vertices.append((x6, y6))

    # And that's one quarter of the bundle outline covered. The remaining vertices can
    # be obtained by changing signs appropriately.
    quarter = vertices[:]
    quarter_reversed = vertices[::-1]
    vertices.extend((v[0], -v[1]) for v in quarter_reversed)
    vertices.extend((-v[0], -v[1]) for v in quarter)
    vertices.extend((-v[0], v[1]) for v in quarter_reversed)

    # So far we have assumed the bundle center to be at (0, 0), so we have to correct
    # this.
    vertex_positions = [translate(center, v) for v in vertices]

    line = LinePathAnnotation(
        vertex_positions, wcs=wcs, closed=True, edgecolor=_OUTLINE_COLOR
    )
    return line


def _science_fiber_outlines(
    r: Angle, d_horizontal: Angle, d_vertical: Angle, center: SkyCoord, wcs: WCS
) -> GroupAnnotation:
    annotation = GroupAnnotation([])
    for row in range(-6, 7):
        if row == -6 or row == 6:
            values = range(-7, 7)
        elif row == -5 or row == 5:
            values = range(-8, 7)
        elif row == -4 or row == 4:
            values = range(-8, 8)
        elif row == -3 or row == 3:
            values = range(-9, 8)
        elif row == -2 or row == 2:
            values = range(-9, 9)
        elif row == -1 or row == 1:
            values = range(-9, 8)
        else:
            values = range(-9, 9)
        for v in values:
            x = (
                (v + 0.5) * d_horizontal.to_value(u.arcsec)
                if row % 2 == 0
                else (v + 1) * d_horizontal.to_value(u.arcsec)
            )
            y = row * d_vertical.to_value(u.arcsec)
            circle = CircleAnnotation(
                center=center, radius=r, wcs=wcs, edgecolor="darkgray", facecolor="none"
            ).translate((x, y) * u.arcsec)
            annotation.add_item(circle)
    return annotation


def _sky_bundle(
    r: Angle,
    d_horizontal: Angle,
    d_vertical: Angle,
    center: SkyCoord,
    wcs: WCS,
    include_fibers: bool,
) -> GroupAnnotation:
    """
    Return an annotation showing the sky bundle.

    The annotation includes the sky bundle outline and, optionally, the bundle fiber
    outlines.

    Parameters
    ----------
    r: `~astropy.coord.Angle`
        The fiber radius, as angle on the sky.
    d_horizontal: `~astropy.coord.Angle`
        The distance between adjacent fibers in the direction of right ascension, as an
        angle on the sky.
    d_vertical: `~astropy.coord.Angle`
        The distance between adjacent rows of fiber cdenters in the direction of
        declination, as an angle on the sky.
    center: `~astropy.coord.SkyCoords`
        The position of the bundle center, as right ascension and declination.
    wcs: `~astropy.wcs.WCS`
        WCS object.
    include_fibers: `bool`
        Whether to include the fiber outlines in the annotation.

    Returns
    -------
    `~imephu.annotation.general.GroupAnnotation`
        The annotation.
    """
    annotation = GroupAnnotation([])
    if include_fibers:
        annotation.add_item(
            _sky_fiber_outlines(
                r=r,
                d_horizontal=d_horizontal,
                d_vertical=d_vertical,
                center=center,
                wcs=wcs,
            )
        )
    annotation.add_item(
        _sky_bundle_outline(
            r=r,
            d_horizontal=d_horizontal,
            d_vertical=d_vertical,
            center=center,
            wcs=wcs,
        )
    )
    return annotation


def _sky_bundle_outline(
    r: Angle, d_horizontal: Angle, d_vertical: Angle, center: SkyCoord, wcs: WCS
) -> LinePathAnnotation:
    # For simplicity, we first assume that the bundle center is located at (0, 0).
    vertices = [
        (-8.5 * d_horizontal - r, r),
        (-7 * d_horizontal - r, r),
        (-7 * d_horizontal - r, d_vertical + r),
        (-4 * d_horizontal + r, d_vertical + r),
        (-4 * d_horizontal + r, r),
        (3 * d_horizontal - r, r),
        (3 * d_horizontal - r, d_vertical + r),
        (8 * d_horizontal + r, d_vertical + r),
        (8 * d_horizontal + r, r),
        (8.5 * d_horizontal + r, r),
    ]

    # And that's one half of the bundle outline covered. The remaining vertices can be
    # obtained by changing signs appropriately.
    half = vertices[:]
    vertices.extend((-v[0], -v[1]) for v in half)

    # So far we have assumed the bundle center to be at (0, 0), so we have to correct
    # this.
    vertex_positions = [translate(center, v) for v in vertices]

    line = LinePathAnnotation(
        vertex_positions, wcs=wcs, closed=True, edgecolor=_OUTLINE_COLOR
    )
    return line


def _sky_fiber_outlines(
    r: Angle, d_horizontal: Angle, d_vertical: Angle, center: SkyCoord, wcs: WCS
) -> GroupAnnotation:
    annotation = GroupAnnotation([])

    for v in list(range(-7, -3)) + list(range(3, 9)):
        x = v * d_horizontal.to_value(u.arcsec)
        y = d_vertical.to_value(u.arcsec)
        circle = CircleAnnotation(
            center=center, radius=r, wcs=wcs, edgecolor="darkgray", facecolor="none"
        ).translate((x, y) * u.arcsec)
        annotation.add_item(circle)

    for v in list(range(-9, 9)):
        x = (v + 0.5) * d_horizontal.to_value(u.arcsec)
        y = 0
        circle = CircleAnnotation(
            center=center, radius=r, wcs=wcs, edgecolor="darkgray", facecolor="none"
        ).translate((x, y) * u.arcsec)
        annotation.add_item(circle)

    for v in list(range(-8, -2)) + list(range(4, 8)):
        x = v * d_horizontal.to_value(u.arcsec)
        y = -d_vertical.to_value(u.arcsec)
        circle = CircleAnnotation(
            center=center, radius=r, wcs=wcs, edgecolor="darkgray", facecolor="none"
        ).translate((x, y) * u.arcsec)
        annotation.add_item(circle)

    return annotation



from typing import List, Sequence

from astropy.coordinates import Angle, SkyCoord
from astropy.visualization.wcsaxes import WCSAxes

from imephu.annotation import Annotation


class GroupAnnotation(Annotation):
    """
    An annotation made of a group of annotations.

    The annotation comprises a list of annotation items. Any of the
    `~impehu.annotation.Annotation` methods are applied to all the items in the list,
    in the order the items appear in the list. So, for example, the ``apply_to`` method
    calls the ``apply_to`` method for all list items, starting from the first item in
    the list and finishing with the last item.

    Annotations can be added to the group when calling the constructor or by using the
    ``add_item`` method afterwards.

    Parameters
    ----------
    items: sequence of `~imephu.annotation.Annotation`
        The initial list of group items.
    """

    def __init__(self, items: Sequence[Annotation]):
        self._items: List[Annotation] = list(items)

    def add_item(self, item: Annotation) -> None:
        """Add an annotation to this group.

        Parameters
        ----------
        item: `~imephu.annotation.Annotation`
            Annotation to add to this group.
        """
        self._items.append(item)

    def add_to(self, ax: WCSAxes) -> None:
        """Add this annotation to a finder chart.

        Parameters
        ----------
        ax: `~astropy.visualization.wcsaxes.WCSAxes`
            Plot axes.
        """
        for item in self._items:
            item.add_to(ax)

    def rotate(self, pivot: SkyCoord, angle: Angle) -> "GroupAnnotation":
        """Rotate this annotation around a pivot and return the result.

        The rotation angle is an angle on the sky, measured from north to east.

        Parameters
        ----------
        pivot: `~astropy.coordinates.SkyCoord`
            Point around which to rotate the annotation.
        angle: `~astropy.coordinates.Angle`
            Angle of rotation, measured from north towards the east.

        Returns
        -------
        `~imephu.annotation.general.GroupAnnotation`
            The annotation resulting from the rotation.
        """
        rotated_items = [item.rotate(pivot, angle) for item in self._items]
        rotated_annotation = GroupAnnotation(items=rotated_items)
        return rotated_annotation

    def translate(self, displacement: Angle) -> "GroupAnnotation":
        """
        Move this annotation along a displacement vector and return the result.

        Parameters
        ----------
        displacement: 2D array of angles
            The displacement by which to move the annotation.

        Returns
        -------
        `~imephu.annotation.general.GroupAnnotation`
            The annotation resulting from the translation.
        """
        translated_items = [item.translate(displacement) for item in self._items]
        translated_annotation = GroupAnnotation(translated_items)
        return translated_annotation



import zipfile
from dataclasses import dataclass
from pathlib import Path
from typing import List

from astropy import units as u
from astropy.coordinates import Angle, SkyCoord
from defusedxml.minidom import parseString


@dataclass()
class MosMaskSlit:
    """A slit in a MOS mask.

    The slit is characterised by its center position, width, height and tilt. The tilt
    should not be confused with the position angle of the MOS mask. The slit will be
    rotated by the sum of the position angle and its tilt.

    Attributes
    ----------
    center: `~astropy.coordinates.SkyCoord`
        The position of the slit center, as right ascension and declination.
    width: `~astropy.coordinates.Angle`
        The slit width, as an angle on the sky.
    height: `~astropy.coordinates.Angle`
        The slit height, as an angle on the sky.
    tilt: `~astropy.coordinates.Angle`
        The slit tilt, as an angle on the sky measured from north to east.
    """

    center: SkyCoord
    width: Angle
    height: Angle
    tilt: Angle


class MosMask:
    """A SALT MOS mask.

    Parameters
    ----------
    mask_xml: str
        XML defining the MOS mask content.

    Attributes
    ----------
    center: `~astropy.coordinates.SkyCoord`
        The position of the mask center, as right ascension and declination.
    position_angle: `~astropy.coordinates.Angle`
        The position angle of the mask.
    reference_stars: list of `astropy.coordinates.SkyCoord`
        The positions of the reference stars, as right ascensions and declinations.
    slits: list of `MosMaskSlit`
        The slits on the mask.
    """

    _center: SkyCoord

    _position_angle: Angle

    _reference_stars: List[SkyCoord]

    _slits: List[MosMaskSlit]

    def __init__(self, mask_xml: str) -> None:
        self._init_from_xml(mask_xml)

    @staticmethod
    def from_file(name: Path) -> "MosMask":
        """Return the `~imephu.salt.utils.MosMask defined in a file.

        The file can be an RSMT file, as produced by SALT's RSS Slit Mask Tool, or an
        XML file.

        Parameters
        ----------
        path: `~pathlib.Path`
            Path of the file defining the MOS mask.
        """
        if zipfile.is_zipfile(name):
            with zipfile.ZipFile(name, "r") as archive:
                mask_xml = archive.read("Slitmask.xml").decode("utf-8")
        else:
            with open(name, "r") as f:
                mask_xml = f.read()

        return MosMask(mask_xml)

    @property
    def center(self) -> SkyCoord:
        """Return the mask center."""
        return self._center

    @property
    def position_angle(self) -> Angle:
        """Return the position angle."""
        return self._position_angle

    @property
    def reference_stars(self) -> List[SkyCoord]:
        """Return the list of reference star positions."""
        return self._reference_stars

    @property
    def slits(self) -> List[MosMaskSlit]:
        """Return the list of slits."""
        return self._slits

    def _init_from_xml(self, mask_xml: str) -> None:
        # Get the DOM
        doc = parseString(mask_xml)

        # Extract the mask position and rotation angle
        parameter_elements = doc.getElementsByTagName("parameter")
        parameters = {}
        for par in parameter_elements:
            name = par.getAttribute("name")
            val = par.getAttribute("value")
            parameters[name] = val
        ra = float(parameters["CENTERRA"]) * u.deg
        dec = float(parameters["CENTERDEC"]) * u.deg
        self._center = SkyCoord(ra=ra, dec=dec)
        self._position_angle = float(parameters["ROTANGLE"]) * u.deg

        # Extract the slit details
        slit_elements = doc.getElementsByTagName("slit")
        self._slits = []
        for slit in slit_elements:
            ra = float(slit.attributes["xce"].value) * u.deg
            dec = float(slit.attributes["yce"].value) * u.deg
            width = float(slit.attributes["width"].value) * u.arcsec
            height = float(slit.attributes["length"].value) * u.arcsec
            tilt = 0.0 * u.deg
            if "tilt" in slit.attributes.keys():
                tilt = float(slit.attributes["tilt"].value) * u.deg
            self._slits.append(
                MosMaskSlit(
                    center=SkyCoord(ra=ra, dec=dec),
                    width=width,
                    height=height,
                    tilt=tilt,
                )
            )

        # Extract the reference star details
        reference_star_elements = doc.getElementsByTagName("refstar")
        self._reference_stars = []
        for ref in reference_star_elements:
            ra = float(ref.attributes["xce"].value) * u.deg
            dec = float(ref.attributes["yce"].value) * u.deg
            self._reference_stars.append(SkyCoord(ra=ra, dec=dec))



from typing import Any, Tuple, Union

import numpy as np
from astropy.coordinates import Angle, SkyCoord
from astropy.wcs import WCS
from numpy import typing as npt

from imephu.annotation.general import GroupAnnotation, LinePathAnnotation
from imephu.geometry import pixel_scales, pixel_to_sky_position, sky_position_to_pixel


class ArrowAnnotation(GroupAnnotation):
    """An annotation for plotting an arrow from a start to an end position.

    The start of the arrow's tail and the tip of the arrow's head are given as sky
    positions (in right ascension and declination). The width and height of the arrow
    head are given as an angle on the sky, and this angle is converted to pixels using
    the pixel scale in right ascension direction.

     By default the arrow is not filled, but you can change this behavior with the
     ``head_filled`` parameter.

    Parameters
    ----------
    start: `~astropy.coordinates.SkyCoord`
        Position of the arrow tail's start.
    end: `~astropy.coordinates.SkyCoord`
        Position of the arrow head's tip.
    wcs: `~astropy.wcs.WCS`
        WCS object.
    head_width: `~astropy.coordinates.Angle`
        Width of the arrow head.
    head_height: `~astropy.coordinates.Angle`
        Height of the arrow head.
    head_filled: `bool`, default: False
        Whether the arrow head is filled.
    tail_shown: `bool`, default: True
        Whether the tail should be shown. If ``False``, only the head is displayed.
    color: color, default: "black"
        The arrow color.
    **kwargs: dict, optional
        Additional keyword arguments, which will be passed to the
        `~imephu.annotation.general.LinePathAnnotation` constructor for both tail and
        head.
    """

    def __init__(
        self,
        start: SkyCoord,
        end: SkyCoord,
        wcs: WCS,
        head_width: Angle,
        head_height: Angle,
        head_filled: bool = False,
        tail_shown: bool = True,
        color: Union[str, Tuple[float, float, float]] = "black",
        **kwargs: Any,
    ):
        super().__init__([])
        if tail_shown:
            super().add_item(ArrowAnnotation._tail(start, end, wcs, color, **kwargs))
        super().add_item(
            ArrowAnnotation._head(
                start, end, wcs, head_width, head_height, head_filled, color, **kwargs
            )
        )

    @staticmethod
    def _tail(
        start: SkyCoord,
        end: SkyCoord,
        wcs: WCS,
        color: Union[str, Tuple[float, float, float]],
        **kwargs: Any,
    ) -> LinePathAnnotation:
        return LinePathAnnotation(
            vertices=[start, end], wcs=wcs, closed=False, edgecolor=color, **kwargs
        )

    @staticmethod
    def _head(
        start: SkyCoord,
        end: SkyCoord,
        wcs: WCS,
        head_width: Angle,
        head_height: Angle,
        head_filled: bool,
        color: Union[str, Tuple[float, float, float]],
        **kwargs: Any,
    ) -> LinePathAnnotation:
        # Let S, be the start of the arrow's tail, T be the tip of the arrow and A, B
        # the other two vertices of the head. Then AA' is bisected by the line ST, and
        # AB and ST are perpendicular to each other.
        # Let's get the unit vector n_TS from head to tail in pixels.
        start_px = sky_position_to_pixel(start, wcs)
        end_px = sky_position_to_pixel(end, wcs)
        end_to_start_px = start_px - end_px  # noqa
        n_TS_px = end_to_start_px / np.linalg.norm(end_to_start_px)

        # We can use the vector to get the point of intersection of AA' and ST.
        sx, _ = pixel_scales(wcs)
        head_height_px = float(head_height / sx)
        point_of_intersection = end_px + head_height_px * n_TS_px  # noqa

        # We can also use it to get a unit vector n_AB along AB.
        n_AB_px: npt.NDArray[np.float_] = np.array([n_TS_px[1], -n_TS_px[0]])

        # And hence we can calculate A and B.
        head_width_px = float(head_width / sx) / 2
        v_A_px = point_of_intersection - head_width_px * n_AB_px  # noqa
        v_B_px = point_of_intersection + head_width_px * n_AB_px  # noqa
        v_A = pixel_to_sky_position(v_A_px, wcs)
        v_B = pixel_to_sky_position(v_B_px, wcs)

        return LinePathAnnotation(
            vertices=[v_A, end, v_B],
            wcs=wcs,
            closed=head_filled,
            edgecolor=color,
            facecolor=color if head_filled else "none",
            **kwargs,
        )



import bisect
from datetime import datetime, timedelta, timezone
from typing import List, Optional

from astropy import units as u
from astropy.coordinates import SkyCoord
from astropy.units import Quantity
from astroquery.jplhorizons import Horizons
from dateutil.parser import parse
from imephu.utils import Ephemeris, MagnitudeRange, SkyCoordRate


class HorizonsService:
    """A service for querying ephemerides from JPL Horizons.

    The service is initialised by calling the constructor with a start and end time,
    and ephemerides can be queried with the `ephemerides` method, which takes a start
    and end time as well. The first time this method is called, JPL Horizons is queried
    for the ephemerides for the interval defined by the _constructor's_ start and end
    time, and the returned list of ephemerides is stored. This list is used to find the
    subset of ephemerides covering the interval defined by the start and end time passed
    to the method. Subsequent calls to the `ephemerides` method use this list as well.

    Effectively this means irrespective how often you call the `ephemerides` method,
    only one request is made to the JPL Horizons service. It also means that the time
    interval passed to the `ephemerides` method must be fully contained within the
    interval passed to the constructor.

    The magnitude range for the ephemerides is given for the V band, and the minimum
    and maximum magnitude are always equal. If the Horizons data contain no V band
    magnitude, no magnitude range is included in the ephemeris.

    The time between ephemeris values is rounded to the nearest minute.

    .. note::
       To ensure full coverage of the interval from ``start`` to ``end``, the actual
       start and end time passed when querying JPL Horizons are slightly different from
       ``start`` and ``end``.

    Parameters
    ----------
    object_id: `str`
        The Horizons identifier for this object, such as ``Ceres``. The exact string
        will be used when querying the ephemerides from JPL Horizons.
    location: `str`
        The location as an `MPC observatory
        code<https://minorplanetcenter.net//iau/lists/ObsCodesF.html>`_ For example, the
        location for the Southern African Large Telescope is B31.
    start: `~datetime.datetime`
        The start of the interval for which ephemerides should be queried. This must be
        a timezone-aware `~datetime.datetime` object.
    end: `~datetime.datetime`
        The end of the interval for which ephemerides should be queried. This must be
        a timezone-aware `~datetime.datetime` object.
    stepsize: `~astropy.units.Quantity`, default: 5 minutes
        The time between ephemeris values. This must be at least 5 minutes. The time is
        rounded to the nearest integer when making the query.
    """

    def __init__(
        self,
        object_id: str,
        location: str,
        start: datetime,
        end: datetime,
        stepsize: Quantity = 5 * u.minute,
    ) -> None:
        # check that the start and end time are timezone-aware
        if start.tzinfo is None or start.tzinfo.utcoffset(None) is None:
            raise ValueError("start must be a timezone-aware datetime.")
        if end.tzinfo is None or end.tzinfo.utcoffset(None) is None:
            raise ValueError("end must be a timezone-aware datetime.")

        # check that the start and end time are consistent
        if start >= end:
            raise ValueError("The start time must be earlier than the end time.")

        # avoid overly excessive queries
        if stepsize < 5 * u.minute:
            raise ValueError(
                "The time between ephemeris values must be at least 5 " "minutes"
            )

        self._object_id = object_id
        self._location = location
        self._start = start
        self._end = end
        self._stepsize = stepsize
        self._ephemerides: Optional[List[Ephemeris]] = None

    def ephemerides(
        self, start: Optional[datetime] = None, end: Optional[datetime] = None
    ) -> List[Ephemeris]:
        """
        Return the list of ephemerides covering an interval.

        The start and time of the interval must be timezone-aware, and the interval must
        be a subinterval of that passed to the constructor. The JPL Horizons service is
        only queried the first time you call this method; subsequent calls use cached
        results.

        If the start or end time is ``None``, the start or end time of the constructor
        is used.

        To ensure full coverage, the last ephemeris epoch returned may be slightly later
        than the end time.

        Parameters
        ----------
        start: `~datetime.datetime`, default: None
            The start time of the interval to cover. This must be a timezone-aware
            `~datetime.datetime` object.
        end: `~datetime.datetime`, default: None
            The end time of the interval to cover. This must be a timezone-aware
            `~datetime.datetime` object.

        Returns
        -------
        list of `~imephu.utils.Ephemeris`
            The list of ephemerides covering the interval from ``start`` to ``end``.
        """
        if start is None:
            start = self._start
        if end is None:
            end = self._end

        # check that the start and end time are timezone-aware
        if start.tzinfo is None or start.tzinfo.utcoffset(None) is None:
            raise ValueError("start must be a timezone-aware datetime.")
        if end.tzinfo is None or end.tzinfo.utcoffset(None) is None:
            raise ValueError("end must be a timezone-aware datetime.")

        # check that the start and end time are consistent
        if start >= end:
            raise ValueError("The start time must be earlier than the end time.")

        # check that the passed interval is a subinterval of that passed to the
        # constructor
        if start < self._start:
            raise ValueError(
                "The start time must not be earlier than that of the " "constructor."
            )
        if end > self._end:
            raise ValueError(
                "The end time must not be later than that of the" "constructor."
            )

        if self._ephemerides is None:
            self._query_horizons()

        return self._cover_time_interval(start, end)

    def _query_horizons(self) -> None:
        # Make sure the whole time interval is covered by the queried ephemerides
        start = self._start.astimezone(timezone.utc).strftime("%Y-%m-%d %H:%M:%S")
        end_time_with_margin = self._end + timedelta(
            seconds=self._stepsize.to_value(u.second)
        )
        stop = end_time_with_margin.astimezone(timezone.utc).strftime(
            "%Y-%m-%d %H:%M:%S"
        )

        # Horizons requires an int for the step size. As round() might call NumPy's
        # round method and thus produce a float, we have to round "manually" using
        # the int function.
        step = f"{int(0.5 + self._stepsize.to_value(u.minute))}m"

        obj = Horizons(
            id=self._object_id,
            location=self._location,
            epochs={"start": start, "stop": stop, "step": step},
        )
        ephemerides = obj.ephemerides()

        # store the ephemerides in the format we need
        self._ephemerides = []
        for row in range(len(ephemerides)):
            epoch = parse(ephemerides["datetime_str"][row]).replace(tzinfo=timezone.utc)
            ra = float(ephemerides["RA"][row]) * u.deg
            dec = float(ephemerides["DEC"][row]) * u.deg
            ra_rate = float(ephemerides["RA_rate"][row]) * u.arcsec / u.hour
            dec_rate = float(ephemerides["DEC_rate"][row]) * u.arcsec / u.hour
            if "V" in ephemerides.keys():
                magnitude = float(ephemerides["V"][row])
                magnitude_range = MagnitudeRange(
                    min_magnitude=magnitude, max_magnitude=magnitude, bandpass="V"
                )
            else:
                magnitude_range = None
            self._ephemerides.append(
                Ephemeris(
                    position=SkyCoord(ra=ra, dec=dec),
                    position_rate=SkyCoordRate(ra=ra_rate, dec=dec_rate),
                    magnitude_range=magnitude_range,
                    epoch=epoch,
                )
            )

    def _cover_time_interval(self, start: datetime, end: datetime) -> List[Ephemeris]:
        """Find the smallest list of ephemerides covering a time interval."""
        # check that the start and end time are timezone-aware
        if start.tzinfo is None or start.tzinfo.utcoffset(None) is None:
            raise ValueError("start must be a timezone-aware datetime.")
        if end.tzinfo is None or end.tzinfo.utcoffset(None) is None:
            raise ValueError("end must be a timezone-aware datetime.")

        # check that JPL Horizons has been queried already
        if self._ephemerides is None:
            raise ValueError("self._ephemerides must not be None.")

        # check that the ephemerides cover the give interval
        if start < self._ephemerides[0].epoch or self._ephemerides[-1].epoch < end:
            raise ValueError("self._ephemerides does not cover the interval.")

        # find the smallest interval covering the time interval
        all_times = [e.epoch for e in self._ephemerides]
        start_index = bisect.bisect_right(all_times, start)
        end_index = bisect.bisect(all_times, end)
        if end_index > 0 and all_times[end_index - 1] == end:
            end_index -= 1

        return self._ephemerides[start_index - 1 : end_index + 1]



from datetime import timezone
from typing import List, Tuple, Union

import numpy as np
from astropy import units as u
from astropy.coordinates import Angle, SkyCoord
from astropy.wcs import WCS

from imephu.annotation.general import (
    ArrowAnnotation,
    CircleAnnotation,
    GroupAnnotation,
    LinePathAnnotation,
    TextAnnotation,
)
from imephu.geometry import (
    pixel_scales,
    pixel_to_sky_position,
    sky_position_to_pixel,
    translate,
)
from imephu.utils import Ephemeris

_MAX_SMALL_MOTION_PIXELS = 18.7


def motion_annotation(ephemerides: List[Ephemeris], wcs: WCS) -> GroupAnnotation:
    """Return an annotation showing the motion of an object on the sky.

    If the motion is sufficiently large, it is shown as an arrow from the initial to the
    final position. Labels for the initial and final epoch are added.

    Otherwise an arrow head pointing in the direction of motion and labels for the
    initial and final epoch are shown. A circle is added around the arrowhead for
    highlighting purposes.

    A sufficiently large motion is a motion where the angular distance between the
    initial and final position is larger than 2 / 75 of the finder chart size.

    Parameters
    ----------
    ephemerides: list of `~imephu.utils.Ephemeris`
        The ephemerides for the motion.
    wcs: `~astropy.wcs.WCS`
        WCS object.

    Returns
    -------
    `~imephu.annotation.general.GroupAnnotation`
        The annotation describing the motion.
    """
    sorted_ephemerides = sorted(ephemerides[:], key=lambda e: e.epoch)
    initial_position = sorted_ephemerides[0].position
    final_position = sorted_ephemerides[-1].position
    distance_travelled = initial_position.separation(final_position)
    if distance_travelled == 0 * u.arcsec:
        raise ValueError("The initial and final position must not be identical.")

    color = "blue"
    if distance_travelled > _MAX_SMALL_MOTION_PIXELS * _angle_per_pixel(wcs):
        return _large_motion_annotation(sorted_ephemerides, wcs, color)
    else:
        return _small_motion_annotation(sorted_ephemerides, wcs, color)


def _large_motion_annotation(
    sorted_ephemerides: List[Ephemeris],
    wcs: WCS,
    color: Union[str, Tuple[float, float, float]],
) -> GroupAnnotation:
    # path showing the motion
    path_annotation = LinePathAnnotation(
        vertices=[e.position for e in sorted_ephemerides],
        wcs=wcs,
        closed=False,
        edgecolor=color,
    )

    # arrow head
    angle_per_pixel = _angle_per_pixel(wcs)
    arrow_annotation = ArrowAnnotation(
        start=sorted_ephemerides[-2].position,
        end=sorted_ephemerides[-1].position,
        wcs=wcs,
        head_width=5.5 * angle_per_pixel,
        head_height=8.4 * angle_per_pixel,
        head_filled=False,
        tail_shown=False,
        color=color,
    )

    # epoch labels
    initial_epoch_label = _large_motion_epoch_label(
        sorted_ephemerides=sorted_ephemerides,
        for_initial_epoch=True,
        color=color,
        wcs=wcs,
    )
    final_epoch_label = _large_motion_epoch_label(
        sorted_ephemerides=sorted_ephemerides,
        for_initial_epoch=False,
        color=color,
        wcs=wcs,
    )

    return GroupAnnotation(
        [path_annotation, arrow_annotation, initial_epoch_label, final_epoch_label]
    )


def _small_motion_annotation(
    sorted_ephemerides: List[Ephemeris],
    wcs: WCS,
    color: Union[str, Tuple[float, float, float]],
) -> GroupAnnotation:
    # mid position on the line between initial and final position on the finder chart,
    # which may be slightly different from the actual mid position on a great circle
    # between the points on the sky
    mid_position = _mid_point_on_chart(sorted_ephemerides, wcs)
    mid_position_px = sky_position_to_pixel(mid_position, wcs)

    # calculate an arrow start and end so that the arrow head will be placed at the
    # right direction and will point in the right direction
    initial_position_px = sky_position_to_pixel(sorted_ephemerides[0].position, wcs)
    final_position_px = sky_position_to_pixel(sorted_ephemerides[-1].position, wcs)
    direction = final_position_px - initial_position_px
    direction_unit = direction / np.linalg.norm(direction)
    arrow_start_px = mid_position_px - (_MAX_SMALL_MOTION_PIXELS / 2) * direction_unit
    arrow_start = pixel_to_sky_position(arrow_start_px, wcs)
    arrow_end_px = mid_position_px + (_MAX_SMALL_MOTION_PIXELS / 2) * direction_unit
    arrow_end = pixel_to_sky_position(arrow_end_px, wcs)

    # arrow head
    angle_per_pixel = _angle_per_pixel(wcs)
    arrow_annotation = ArrowAnnotation(
        start=arrow_start,
        end=arrow_end,
        wcs=wcs,
        head_width=15 * angle_per_pixel,
        head_height=15 * angle_per_pixel,
        head_filled=False,
        tail_shown=False,
        color=color,
    )

    # circle for highlighting the arrow
    circle_annotation = CircleAnnotation(
        center=mid_position,
        radius=(_MAX_SMALL_MOTION_PIXELS / 2 + 5) * angle_per_pixel,
        wcs=wcs,
        edgecolor=color,
    )

    # epoch labels
    initial_epoch_label = _small_motion_epoch_label(
        sorted_ephemerides=sorted_ephemerides,
        for_initial_epoch=True,
        color=color,
        wcs=wcs,
    )
    final_epoch_label = _small_motion_epoch_label(
        sorted_ephemerides=sorted_ephemerides,
        for_initial_epoch=False,
        color=color,
        wcs=wcs,
    )

    return GroupAnnotation(
        [arrow_annotation, circle_annotation, initial_epoch_label, final_epoch_label]
    )


def _large_motion_epoch_label(
    sorted_ephemerides: List[Ephemeris],
    for_initial_epoch: bool,
    color: Union[str, Tuple[float, float, float]],
    wcs: WCS,
) -> TextAnnotation:
    # For the initial epoch the label should be put slightly to the south if the object
    # is moving north.
    angle_per_pixel = _angle_per_pixel(wcs)
    initial_dec_displacement = 12 * angle_per_pixel
    if sorted_ephemerides[1].position.dec >= sorted_ephemerides[0].position.dec:
        initial_dec_displacement *= -1

    # The final epoch label should be displaced in the opposite direction as compared to
    # the initial epoch one.
    final_dec_displacement = -initial_dec_displacement

    dec_displacement = (
        initial_dec_displacement if for_initial_epoch else final_dec_displacement
    )
    dec_displacement_arcsec = dec_displacement.to_value(u.arcsec)
    if for_initial_epoch:
        ephemeris = sorted_ephemerides[0]
    else:
        ephemeris = sorted_ephemerides[-1]
    position = translate(ephemeris.position, (0, dec_displacement_arcsec) * u.arcsec)
    text = ephemeris.epoch.astimezone(timezone.utc).strftime("%Y-%m-%d %H:%M UT")

    return TextAnnotation(position=position, s=text, wcs=wcs, color=color, fontsize=8)


def _small_motion_epoch_label(
    sorted_ephemerides: List[Ephemeris],
    for_initial_epoch: bool,
    color: Union[str, Tuple[float, float, float]],
    wcs: WCS,
) -> TextAnnotation:
    # For the initial epoch the label should be put slightly to the south if the object
    # is moving north.
    angle_per_pixel = _angle_per_pixel(wcs)
    initial_dec_displacement = (_MAX_SMALL_MOTION_PIXELS / 2 + 13) * angle_per_pixel
    if sorted_ephemerides[1].position.dec >= sorted_ephemerides[0].position.dec:
        initial_dec_displacement *= -1

    # The final epoch label should be displaced in the opposite direction as compared to
    # the initial epoch one.
    final_dec_displacement = -initial_dec_displacement

    dec_displacement = (
        initial_dec_displacement if for_initial_epoch else final_dec_displacement
    )
    dec_displacement_arcsec = dec_displacement.to_value(u.arcsec)
    if for_initial_epoch:
        ephemeris = sorted_ephemerides[0]
    else:
        ephemeris = sorted_ephemerides[-1]
    mid_point = _mid_point_on_chart(sorted_ephemerides, wcs)
    position = translate(mid_point, (0, dec_displacement_arcsec) * u.arcsec)
    text = ephemeris.epoch.astimezone(timezone.utc).strftime("%Y-%m-%d %H:%M UT")

    return TextAnnotation(position=position, s=text, wcs=wcs, color=color, fontsize=8)


def _angle_per_pixel(wcs: WCS) -> Angle:
    """Return the angle per pixel for a WCS object.

    The angle per pixel is averaged over right ascension and declination.
    """
    sx, sy = pixel_scales(wcs)
    return (sx + sy) / 2


def _mid_point_on_chart(sorted_ephemerides: List[Ephemeris], wcs: WCS) -> SkyCoord:
    """Return the midpoint between the initial and final position.

    The midpoint is the point on the sky which on the finder chart corresponds to the
    midpoint of the line between the initial and final position. This in general may be
    different from the midpoint on the great circle through the initial and final
    positions on the sky.
    """
    initial_position_px = sky_position_to_pixel(sorted_ephemerides[0].position, wcs)
    final_position_px = sky_position_to_pixel(sorted_ephemerides[-1].position, wcs)
    mid_position_px = (initial_position_px + final_position_px) / 2
    return pixel_to_sky_position(mid_position_px, wcs)



from dataclasses import dataclass
from datetime import datetime
from typing import List, Optional

from astropy.coordinates import SkyCoord
from astropy.units import Quantity


@dataclass
class SkyCoordRate:
    """A rate of change in position on the sky.

    Attributes
    ----------
    ra: `~astropy.unit.Quantity`
        The rate of change in right ascension.
    dec: `~astropy.unit.Quantity`
        The rate of change in declination.
    """

    ra: Quantity

    dec: Quantity


@dataclass
class MagnitudeRange:
    """A magnitude range.

    Attributes
    ----------
    bandpass: `str`
        The bandpass for which the magnitudes are given.
    min_magnitude: `float`
        The minimum (brightest) magnitude.
    max_magnitude: `float`
        The maximum (faintest) magnitude.
    """

    bandpass: str
    min_magnitude: float
    max_magnitude: float


@dataclass
class Ephemeris:
    """An ephemeris with an epoch, position and magnitude range.

    Parameters
    ----------
    epoch: `~datetime.datetime`
        The epoch, i.e. the datetime for which the position is given. The epoch must be
        a timezone-aware datetime.
    position: `~astropy.coordinates.SkyCoord`
        The position, in right ascension and declination.
    position_rate: SkyCoordRate
        The rate of change in the position on the sky.
    magnitude_range: `~imephu.utils.MagnitudeRange`
        The magnitude range.
    """

    epoch: datetime

    position: SkyCoord

    position_rate: SkyCoordRate

    magnitude_range: Optional[MagnitudeRange]

    def __post_init__(self) -> None:
        """Check that the epoch is timezone-aware."""
        if self.epoch.tzinfo is None or self.epoch.tzinfo.utcoffset(None) is None:
            raise ValueError("The epoch must be a timezone-aware datetime object.")


def mid_position(start: SkyCoord, end: SkyCoord) -> SkyCoord:
    """Return the mid position between a start and end position on the sky.

    The mid position is the mid position on the great arc between the start and end
    position.

    Taken from https://github.com/astropy/astropy/issues/5766.

    Parameters
    ----------
    start: `~astropy.coordinates.SkyCoords`
        Start position on the sky.
    end: `~astropy.coordinates.SkyCoords`
        End position on the sky.
    """
    pa = start.position_angle(end)
    separation = start.separation(end)
    return start.directional_offset_by(pa, separation / 2)


def ephemerides_magnitude_range(
    ephemerides: List[Ephemeris],
) -> Optional[MagnitudeRange]:
    """Return the magnitude range for a list of ephemerides.

    The minimum (maximum) magnitude is the minimum (maximum) magnitude for all
    ephemerides. If none of the ephemerides has a magnitude range, ``None`` is returned.

    Parameters
    ----------
    ephemerides: list of `~imephu.utils.Ephemeris`
        The list of ephemerides.

    Returns
    -------
    `~imephu.utils.MagnitudeRange`, optional
        The magnitude range for the list of ephemerides.
    """
    min_magnitude: Optional[float] = None
    max_magnitude: Optional[float] = None
    bandpass: Optional[str] = None
    for ephemeris in ephemerides:
        if ephemeris.magnitude_range:
            if (
                min_magnitude is None
                or ephemeris.magnitude_range.min_magnitude < min_magnitude
            ):
                min_magnitude = ephemeris.magnitude_range.min_magnitude
            if (
                max_magnitude is None
                or ephemeris.magnitude_range.max_magnitude > max_magnitude
            ):
                max_magnitude = ephemeris.magnitude_range.max_magnitude
            if bandpass is None:
                bandpass = ephemeris.magnitude_range.bandpass
            elif ephemeris.magnitude_range.bandpass != bandpass:
                raise ValueError("The bandpass must be the same for all ephemerides.")

    if min_magnitude is not None and max_magnitude is not None and bandpass is not None:
        return MagnitudeRange(
            min_magnitude=min_magnitude, max_magnitude=max_magnitude, bandpass=bandpass
        )
    else:
        return None
